﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Threading;
using System.Reflection;

namespace NauPlaying
{
	/// <summary>
	/// アプリケーション監視プラグインマネージャクラス（シングルトン）
	/// </summary>
	public class AppMonitorPluginManager
	{
		/// <summary>
		/// 唯一のインスタンス
		/// </summary>
		private static AppMonitorPluginManager _instance = null;

		/// <summary>
		/// プレーヤーサーチ優先順位
		/// (nullまたは空配列でサーチしない)
		/// </summary>
		private static PlayerType[] _searchPlayerPriority;
		/// <summary>
		/// プレーヤーサーチ優先順位の設定、取得
		/// (nullまたは空配列でサーチしない)
		/// </summary>
		public PlayerType[] SearchPlayerPriority
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);
					return _searchPlayerPriority;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
			set
			{
				try
				{
					_lockObj.AcquireWriterLock(Timeout.Infinite);
					_searchPlayerPriority = value;
				}
				finally
				{
					_lockObj.ReleaseWriterLock();
				}
			}
		}

		/// <summary>
		/// ロック用オブジェクト
		/// </summary>
		private static ReaderWriterLock _lockObj = new ReaderWriterLock();
		/// <summary>
		/// プラグイン用アプリケーションドメイン
		/// </summary>
		private static AppDomain _appDomain = null;
		/// <summary>
		/// アプリケーション監視プラグイン
		/// </summary>
		private static PluginBase.IAppMonitorPlugin _plugin = null;
		/// <summary>
		/// 現在のプレーヤータイプ
		/// </summary>
		private static PlayerType _playerType = NauPlaying.PlayerType.Unknwon;

		/// <summary>
		/// アプリケーション監視プラグインの取得
		/// (通常はラップされたメソッドを使用してください。ラップされたメソッドはスレッドセーフです。)
		/// </summary>
		public static PluginBase.IAppMonitorPlugin AppMonitor
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);
					return _plugin;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// プラグインが有効かを調べます
		/// </summary>
		public static bool IsValid
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);
					return _plugin != null;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}
		/// <summary>
		/// 現在のプレーヤータイプを取得
		/// </summary>
		public static PlayerType PlayerType
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);
					return _playerType;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// インスタンスの取得
		/// </summary>
		public static AppMonitorPluginManager Instance
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);
					return _instance;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// スタティックコンストラクタ
		/// </summary>
		static AppMonitorPluginManager()
		{
			try
			{
				_lockObj.AcquireWriterLock(Timeout.Infinite);
				if(_instance == null)
					_instance = new AppMonitorPluginManager();
			}
			finally
			{
				_lockObj.ReleaseWriterLock();
			}
		}
		/// <summary>
		/// シングルトンを強制するためのプライベートコンストラクタ
		/// </summary>
		private AppMonitorPluginManager()
		{
		}
		/// <summary>
		/// デストラクタ
		/// </summary>
		~AppMonitorPluginManager()
		{
			Dispose();
		}
		/// <summary>
		/// リソースの開放
		/// </summary>
		public static void Dispose()
		{
			try
			{
				_lockObj.AcquireWriterLock(Timeout.Infinite);

				// アプリケーションモニターの開放
				try
				{
					if(_plugin != null)
					{
						_plugin.Exit(null);
						_plugin.Dispose();
					}
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}
				finally
				{
					_plugin = null;
				}

				// アプリケーションドメインの開放
				try
				{
					if(_appDomain != null)
						AppDomain.Unload(_appDomain);
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}
				finally
				{
					_appDomain = null;
				}

				// プレーヤタイプを初期化
				_playerType = NauPlaying.PlayerType.Unknwon;
			}
			finally
			{
				_lockObj.ReleaseWriterLock();
			}
		}

		/// <summary>
		/// プラグインのリロード
		/// （すでに読み込まれていればそれを開放して。新たにロードします。）
		/// </summary>
		/// <param name="playerType">使用するプレーヤタイプ</param>
		/// <param name="forced">true で終了待ちしないで強制的に起動</param>
		/// <returns></returns>
		public static bool ReloadPlugin(PlayerType playerType, bool forced)
		{
			Dispose();

			try
			{
				_lockObj.AcquireWriterLock(Timeout.Infinite);

				// アプリケーションドメインの生成
				try
				{
					AppDomainSetup setup = new AppDomainSetup();
					setup.ApplicationName = "NauPlaying_Plugin";
					setup.ApplicationBase = AppDomain.CurrentDomain.BaseDirectory;
					setup.PrivateBinPath = Generic.PluginFoldername;

					setup.ShadowCopyFiles = "false";

					// シャドウコピーしたいならコメントアウトをはずす
					/*
					setup.CachePath = Generic.RemoteCachePath;
					setup.ShadowCopyFiles = "true";
					setup.ShadowCopyDirectories = pluginDirectory;
					*/

					_appDomain = AppDomain.CreateDomain("Plugins", null, setup);
					if(_appDomain == null)
					{
						_playerType = NauPlaying.PlayerType.Unknwon;
						return false;
					}
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);

					_playerType = NauPlaying.PlayerType.Unknwon;
					return false;
				}

				// プラグインのロード

				#region /* プラグインのロード中にエラーが発生したときの後始末メソッド loadAsm_ErrorProc */
				Action loadAsm_ErrorProc = () =>
				{
					try
					{
						if(_appDomain != null)
							AppDomain.Unload(_appDomain);
					}
					catch(Exception ex)
					{
						Generic.UpdateLogFile(ex);
					}
					finally
					{
						_appDomain = null;
					}

					_playerType = NauPlaying.PlayerType.Unknwon;
				};
				#endregion

				try
				{
					string assemblyName;
					string className;
					if(playerType == PlayerType.iTunes)
					{
						assemblyName = Generic.iTunesMonitorPluginAssemblyName;
						className = Generic.iTunesMonitorPluginClassName;
					}
					else
					{
						assemblyName = Generic.WinampMonitorPluginAssemblyName;
						className = Generic.WinampMonitorPluginClassName;
					}
				//	PluginBase.IAppMonitorPlugin _plugin = (PluginBase.IAppMonitorPlugin)_appDomain.CreateInstanceAndUnwrap(assemblyName, className);
					
					PluginBase.ISingletonHelper singletonHelper = (PluginBase.ISingletonHelper)_appDomain.CreateInstanceAndUnwrap(assemblyName, Generic.SingletonHelperClassName);
					if(singletonHelper == null)
					{
						loadAsm_ErrorProc();
						return false;
					}
					


					_plugin = (PluginBase.IAppMonitorPlugin)singletonHelper.Instance;
					_plugin.Init(null);
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
					loadAsm_ErrorProc();
					return false;
				}

				// リロードの場合はプレーヤの終了を待機してモニター開始（多重起動の防止）
				try
				{
					if(forced)
						_plugin.StartMonitor(1000, 0);
					else
						_plugin.StartMonitor(1000, 10000);

					_playerType = playerType;
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
					Dispose();
					return false;
				}

				return true;
			}
			finally
			{
				_lockObj.ReleaseWriterLock();
			}
		}

		/// <summary>
		/// 現在読み込まれているプラグインと違う場合、指定のプラグインの読み込みを試みます。
		/// 失敗した場合は、それまで選択されていたプラグインをリロードします。
		/// </summary>
		/// <param name="playerType">使用するプレーヤタイプ</param>
		/// <param name="forced">true で終了待ちしないで強制的に起動</param>
		/// <returns></returns>
		public static bool SwitchPlugin(PlayerType playerType, bool forced)
		{
			PlayerType oldPlayerType = PlayerType;

			if(oldPlayerType == playerType)
				return true;
			
			if(ReloadPlugin(playerType, forced))
				return true;

			ReloadPlugin(oldPlayerType, true);
			return false;
		}

		private static void _searchPlayerTimer_Task(object StateObj)
		{
		}

		#region /* IAppMonitorPlugin のラップメソッド */

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：現在再生中の曲情報を取得します（再生されていない、曲を取得できなかった場合は null）
		/// （プラグインが読み込まれていなたっかり、プレーヤーと接続されていない場合は必ず失敗し null を返します。）
		/// </summary>
		public static PluginBase.TrackInfo CurrentTrackInfo
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return null;

					if(!_plugin.IsConnected)
						return null;

					return _plugin.CurrentTrackInfo;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：現在再生中かを調べる
		/// （プラグインが読み込まれていなたっかり、プレーヤーと接続されていない場合は必ず失敗し false を返します。）
		/// </summary>
		public static bool IsPlaying
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return false;

					if(!_plugin.IsConnected)
						return false;

					return _plugin.IsPlaying;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：再生位置の取得、設定
		/// （プラグインが読み込まれていなたっかり、プレーヤーと接続されていない場合は必ず失敗し、get の場合 -1 を返します。set の場合は何もしません。）
		/// </summary>
		public static int PlayerPosition
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return -1;

					if(!_plugin.IsConnected)
						return -1;

					return _plugin.PlayerPosition;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
			set
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return;

					if(!_plugin.IsConnected)
						return;

					_plugin.PlayerPosition = value;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：再生ボリュームを 0 ～ 100 [%] で取得、設定
		/// （プラグインが読み込まれていなたっかり、プレーヤーと接続されていない場合は必ず失敗し、get の場合 -1 を返します。set の場合は何もしません。）
		/// </summary>
		public static int SoundVolume
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return -1;

					if(!_plugin.IsConnected)
						return -1;

					return _plugin.SoundVolume;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
			set
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return;

					if(!_plugin.IsConnected)
						return;

					_plugin.SoundVolume = value;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：プラグインがプレーヤーと接続しているかを調べます。
		/// （プラグインが読み込まれていない場合は必ず失敗し false を返します。）
		/// </summary>
		public static bool IsConnected
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return false;

					return _plugin.IsConnected;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：プレーヤーを制御します。
		/// （プラグインが読み込まれていなたっかり、プレーヤーと接続されていない場合は必ず失敗し false を返します。）
		/// </summary>
		/// <param name="type">制御タイプ</param>
		/// <returns></returns>
		public static bool ControlPlayer(PluginBase.PlayerControl type)
		{
			try
			{
				_lockObj.AcquireReaderLock(Timeout.Infinite);

				if(_plugin == null)
					return false;

				if(!_plugin.IsConnected)
					return false;

				return _plugin.ControlPlayer(type);
			}
			finally
			{
				_lockObj.ReleaseReaderLock();
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：モニターイベントをポップする
		/// （プラグインが読み込まれていない場合は必ず失敗し false を返します。）
		/// </summary>
		/// <returns></returns>
		public static bool PopAppMonitorEvent(out PluginBase.AppMonitorEvent appEvent)
		{
			try
			{
				_lockObj.AcquireReaderLock(Timeout.Infinite);

				appEvent = PluginBase.AppMonitorEvent.None;

				if(_plugin == null)
					return false;

				appEvent = _plugin.PopAppMonitorEvent();

				return appEvent != PluginBase.AppMonitorEvent.None;
			}
			finally
			{
				_lockObj.ReleaseReaderLock();
			}
		}

		/// <summary>
		/// IAppMonitorPlugin のラップメソッド：最後のエラーを返す（1回もエラーが発生していないときは null）
		/// （プラグインが読み込まれていない場合は必ず null を返します。）
		/// </summary>
		public static Exception LastException
		{
			get
			{
				try
				{
					_lockObj.AcquireReaderLock(Timeout.Infinite);

					if(_plugin == null)
						return null;

					return _plugin.LastException;
				}
				finally
				{
					_lockObj.ReleaseReaderLock();
				}
			}
		}

		#endregion
	}
}
