﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using Win32SDK;

namespace NauPlaying
{
	/// <summary>
	/// ボード表示クラス
	/// </summary>
	public partial class FormBoard : FormLayeredWindow
	{
		/// <summary>
		/// アプリケーション設定
		/// （この設定は Initialise されたときにディープコピーされたもの）
		/// </summary>
		private AppSetting _appSetting;
		/// <summary>
		/// スキン設定
		/// （この設定は Initialise されたときにディープコピーされたもの）
		/// </summary>
		private Skin.SkinSetting _skinSetting;

		/// <summary>
		/// スキンレンダー
		/// </summary>
		private Skin.SkinRender _skinRender;

		/// <summary>
		/// 描画の際に設定された描画設定
		/// (再描画の際に設定される描画情報)
		/// </summary>
		private Skin.DrawingInfo _drawInfo;
		/// <summary>
		/// ボードデータ
		/// </summary>
		private BoardData _boardData;
		/// <summary>
		/// アニメーションコンテナ
		/// </summary>
		private List<Skin.Animation> _animations;

		/// <summary>
		/// 初期化終了時に true。（アニメーション実行可能である）
		/// </summary>
		public bool IsInitialized { private set; get; }
		/// <summary>
		/// アニメーション中かを調べる
		/// </summary>
		public bool IsAnimation { get { return _timer.Enabled; } }

		/// <summary>
		/// null 以外を設定することでカーソルを指定
		/// </summary>
		public Cursor CurrentCursor { set; get; }

		/// <summary>
		/// アニメーション開始イベント
		/// </summary>
		public event EventHandler<EventArgs> OnBeginAnimationEvent;
		/// <summary>
		/// アニメーション終了イベント
		/// </summary>
		public event EventHandler<EventArgs> OnEndAnimationEvent;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		public FormBoard()
		{
			IsInitialized = false;
			_skinRender = new Skin.SkinRender();
			CurrentCursor = null;

			InitializeComponent();
		}

		/// <summary>
		/// ウインドウメッセージのトラップ
		/// </summary>
		/// <param name="m"></param>
		protected override void WndProc(ref Message m)
		{
			// クライアント領域外のマウスメッセージの補足でカーソル形状を制御
			if(m.Msg == Win32.WM_NCMOUSEMOVE || m.Msg == Win32.WM_NCLBUTTONDOWN)
			{
				if(CurrentCursor != null)
					System.Windows.Forms.Cursor.Current = CurrentCursor;
			}

			if(m.Msg == Win32.WM_NCLBUTTONDBLCLK)
			{
				DialogResult = System.Windows.Forms.DialogResult.OK;
				Close();

				m.Result = (IntPtr)0;
				return;
			}

			if(m.Msg == Win32.WM_NCRBUTTONDOWN)
			{
				DialogResult = System.Windows.Forms.DialogResult.Cancel;
				Close();

				m.Result = (IntPtr)0;
				return;
			}

			base.WndProc(ref m);

			// どこをつかんでも移動できるようにする
			if((m.Msg == Win32.WM_NCHITTEST) && (m.Result.ToInt32() == User32.HTCLIENT))
			{
				m.Result = (IntPtr)User32.HTCAPTION;
			}
		}

		/// <summary>
		/// キーイベントをトラップして終了可能にする
		/// </summary>
		/// <param name="keyData"></param>
		/// <returns></returns>
		protected override bool ProcessDialogKey(Keys keyData)
		{
			switch(keyData)
			{
			case Keys.Escape:
				DialogResult = System.Windows.Forms.DialogResult.Cancel;
				Close();
				return false;

			case Keys.Return:
				DialogResult = System.Windows.Forms.DialogResult.OK;
				Close();
				return false;
			}
			return true;
		}

		/// <summary>
		/// フォームが表示される前の処理
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void FormBoard_Load(object sender, EventArgs e)
		{
		}

		/// <summary>
		/// フォームが閉じられるときの処理
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void FormBoard_FormClosing(object sender, FormClosingEventArgs e)
		{
			if(_timer.Enabled)
				_timer.Stop();

			_skinRender.Dispose();
			_skinRender = null;

			System.Diagnostics.Debug.WriteLine("FormBoard_FormClosing!!");
		}

		/// <summary>
		/// ボードの初期化（フォーム表示前に必ず実行すること）
		/// </summary>
		/// <param name="appSetting">アプリケーション設定（ディープコピーされる）</param>
		/// <param name="skinSettings">スキン設定（ディープコピーされる）</param>
		/// <param name="data">ボードデータ（ディープコピーされる）</param>
		public void InitializeBoard(AppSetting appSetting, Skin.SkinSetting skinSettings, BoardData data)
		{
			// 描画設定の保存
			_appSetting = appSetting.DeepCopy();
			_skinSetting = skinSettings.DeepCopy();
			_boardData = data.DeepCopy();

			_drawInfo = new Skin.DrawingInfo();

			// スキンの初期化
			_skinRender.InitilizeBitmaps(_appSetting, _skinSetting);

			if(_skinSetting.EnableArtwork)
				_skinRender.SetArtwork(data.Artwork);

			_skinRender.SetText(data.Text);

			// アニメーション初期化
			Skin.SkinSetting sset = _skinSetting;
			AppSetting aset = _appSetting;

			_timer.Interval = aset.StepTime;

			_animations = new List<Skin.Animation>();

			if(aset.EnablePlayerBoardLocation)
			{
				switch(aset.PlayerType)
				{
				case PlayerType.iTunes:
					_animations.Add(new Skin.ChangeLocationAnimation(aset.iTunesBoardLocation));
					break;

				case PlayerType.Winamp:
					_animations.Add(new Skin.ChangeLocationAnimation(aset.WinampBoardLocation));
					break;

				default:
					_animations.Add(new Skin.ChangeLocationAnimation(aset.BoardLocation));
					break;
				}
			}
			else
				_animations.Add(new Skin.ChangeLocationAnimation(aset.BoardLocation));

			if(aset.FadeinTime > 0)
				_animations.Add(new Skin.FadeAnimation(aset.FadeinTime, aset.StepTime, true));
			else
				_animations.Add(new Skin.ConstAlphaAnimation(1.0f));

			switch(aset.ScrollMode)
			{
			case ScrollMode.Mode1:
				if(_skinRender.IsScrollText)
				{
					Skin.ScrollAnimation animation = new Skin.ScrollAnimation(_skinSetting.TextDispRect.Width, _skinRender.TextDrawingSize.Width, aset.ScrollStep, aset.ScrollAcceleration, true);
					_animations.Add(animation);
					_animations.Add(new Skin.StopAnimation(aset.ScrollStopTime, aset.StepTime));

					// スクロール時間がスクロールしない場合よりも短い場合停止を入れる
					int scrollTime = animation.ScrollCount * aset.StepTime + aset.ScrollStopTime;
					if(scrollTime < aset.NoneScrollStopTime)
						_animations.Add(new Skin.StopAnimation(aset.NoneScrollStopTime - scrollTime, aset.StepTime));
				}
				else
					_animations.Add(new Skin.StopAnimation(aset.NoneScrollStopTime, aset.StepTime));
				break;

			case ScrollMode.Mode2:
				if(_skinRender.IsScrollText)
				{
					Skin.ScrollAnimation animation1 = new Skin.ScrollAnimation(_skinSetting.TextDispRect.Width, _skinRender.TextDrawingSize.Width, aset.ScrollStep, aset.ScrollAcceleration, true);
					Skin.ScrollAnimation animation2 = new Skin.ScrollAnimation(_skinSetting.TextDispRect.Width, _skinRender.TextDrawingSize.Width, aset.ScrollStep, aset.ScrollAcceleration, false);

					_animations.Add(animation1);
					_animations.Add(new Skin.StopAnimation(aset.ScrollStopTime, aset.StepTime));
					_animations.Add(animation2);
					_animations.Add(new Skin.StopAnimation(aset.ScrollStopTime, aset.StepTime));

					// スクロール時間がスクロールしない場合よりも短い場合停止を入れる
					int scrollTime = (animation1.ScrollCount + animation2.ScrollCount) * aset.StepTime + aset.ScrollStopTime * 2;
					if(scrollTime < aset.NoneScrollStopTime)
						_animations.Add(new Skin.StopAnimation(aset.NoneScrollStopTime - scrollTime, aset.StepTime));
				}
				else
					_animations.Add(new Skin.StopAnimation(aset.NoneScrollStopTime, aset.StepTime));
				break;

			default:
				_animations.Add(new Skin.StopAnimation(aset.ScrollStopTime, aset.NoneScrollStopTime));
				break;
			}

			if(aset.FadeoutTime > 0)
				_animations.Add(new Skin.FadeAnimation(aset.FadeoutTime, aset.StepTime, false));

			_animations.Add(new Skin.ConstAlphaAnimation(0.0f));

			IsInitialized = true;
		}

		/// <summary>
		/// アニメーションの開始
		/// </summary>
		public void StartAnimation()
		{
			if(!IsInitialized)
				return;
			if(IsAnimation)
				return;

			IsInitialized = false;

			if(OnBeginAnimationEvent != null)
				OnBeginAnimationEvent(this, new EventArgs());

			_timer.Start();
		}

		/// <summary>
		/// アニメーションの停止
		/// 
		/// （OnEndAnimationEvent デリゲートは処理されない）
		/// </summary>
		public void StopAnimation()
		{
			if(!IsAnimation)
				return;

			_timer.Stop();
		}

		/// <summary>
		/// スキンのプレビュー表示を行う
		/// </summary>
		/// <param name="alpha">アルファ</param>
		/// <param name="location">表示位置</param>
		public void Preview(float alpha, Point location)
		{
			Skin.ConstAlphaAnimation anim = new Skin.ConstAlphaAnimation(alpha);

			Skin.AnimationData args = new Skin.AnimationData();
			args.DrawingInfo = new Skin.DrawingInfo();

			args.DrawingInfo.Alpha = alpha;
			args.DrawingInfo.Location = location;

			anim.Execute(args);

			if(args.IsRedraw)
			{
				Location = args.DrawingInfo.Location;
				_skinRender.UpdateBoard(args.DrawingInfo);
				SetLayeredWindow(_skinRender.LayerBitmap);
			}
		}

		/// <summary>
		/// アニメーション処理
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _timer_Tick(object sender, EventArgs e)
		{
			// アニメーションの実行
			Skin.AnimationData args = new Skin.AnimationData();
			args.DrawingInfo = _drawInfo;

			try
			{
				_animations[0].Execute(args);

				if(args.IsRedraw)
				{
					Location = _drawInfo.Location;
					_skinRender.UpdateBoard(_drawInfo);
					SetLayeredWindow(_skinRender.LayerBitmap);
				}
			}
			catch(Exception ex)
			{
				_animations.Clear();
				_timer.Stop();

				MessageBox.Show("アニメーション実行中に例外が発生しました。" + ex.Message, Generic.AppTitle);

				if(OnEndAnimationEvent != null)
					OnEndAnimationEvent(this, new EventArgs());

				return;
			}

			// 1つのアニメーションが終了した場合の処理
			if(args.IsFinish)
			{
				_animations.RemoveAt(0);

				// すべて終了した場合
				if(_animations.Count <= 0)
				{
					_timer.Stop();

					if(OnEndAnimationEvent != null)
						OnEndAnimationEvent(this, new EventArgs());

					return;
				}
			}
		}
	}
}
