﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;

using Win32SDK;

namespace NauPlaying
{
	public partial class FormMain : Form
	{
		/// <summary>
		/// プラグインイベント監視タイマーの多重実行防止フラグ
		/// </summary>
		private static object _pluginEventTimer_lockObj = new object();
		/// <summary>
		/// イントロタイマーの多重実行防止フラグ
		/// </summary>
		private static object _introTimer_lockObj = new object();
		/// <summary>
		/// プレイヤー検出タイマーの多重実行防止フラグ
		/// </summary>
		private static object _appSearchTimer_lockObj = new object();
		/// <summary>
		/// 同期操作用ロックオブジェクト（必要か？）
		/// </summary>
		private static object _syncObj = new object();

		/// <summary>
		/// アプリケーション設定
		/// </summary>
		private static AppSetting _appSetting = null;
		/// <summary>
		/// スキン設定
		/// </summary>
		private static Skin.SkinSetting _skinSetting = null;

		/// <summary>
		/// イントロ再生時の次回曲送り時間[s]
		/// </summary>
		private int _intorChangeNextTime = 0;
		/// <summary>
		/// イントロ再生時のフェードインフラグ
		/// </summary>
		private bool _introFadeinFlag = false;
		/// <summary>
		/// イントロ再生時のフェードアウトフラグ
		/// </summary>
		private bool _introFadeoutFlag = false;
		/// <summary>
		/// イントロフェード中であれば true
		/// </summary>
		private bool IsIntroFeading
		{
			get { return _introFadeoutFlag | _introFadeinFlag; }
		}
		/// <summary>
		/// フェード動作開始前のユーザー設定ボリューム
		///  (イントロ設定時と、イントロ自にスキップされた際に設定)
		/// </summary>
		private int _introFadeUserVolume = 100;
		/// <summary>
		/// フェード動作に使用する現在の設定ボリューム
		/// </summary>
		private int _introFadeCurrentVolume = 0;
		/// <summary>
		/// フェード動作時のボリューム変化量
		/// </summary>
		private int _introFadeVolumeDelta = 1;

		/// <summary>
		/// イントロ再生モードが有効かどうかの設定
		/// （プロパティー EnableIntroPlayMode を使用する）
		/// </summary>
		private static bool _enableIntroPlayMode = false;
		/// <summary>
		/// イントロ再生モードが有効かどうかの設定
		/// </summary>
		private bool EnableIntroPlayMode
		{
			set
			{
				_enableIntroPlayMode = value;

				if(_enableIntroPlayMode)
				{
					UpdateIntroVolumeSetting();
					UpdateIntroFeadSettting(true);

					_timer_Intro.Start();
				}
				else
				{
					_timer_Intro.Stop();
				}

				enableIntroModeToolStripMenuItem.Checked = _enableIntroPlayMode;
				UpdateNotifyIconImage();
			}
			get { return _enableIntroPlayMode; }
		}

		/// <summary>
		/// トラック変更通知が有効かどうか
		/// </summary>
		private static bool _enableChangeTrackNotify;
		/// <summary>
		/// 強制プレイヤータイプ（Unknwonで強制なし）
		/// </summary>
		private static PlayerType _forcedPlayerType = PlayerType.Unknwon;
		/// <summary>
		/// ボードフォーム
		/// </summary>
		private static FormBoard _formBoard = null;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		public FormMain()
		{
			InitializeComponent();
		}

		/// <summary>
		/// ウインドウメッセージのトラップ
		/// </summary>
		/// <param name="m"></param>
		protected override void WndProc(ref Message m)
		{
			// Winadow 終了時にソフトを終了
			if(m.Msg == Win32.WM_ENDSESSION)
			{
				Close();

				m.Result = IntPtr.Zero;
				return;
			}

			base.WndProc(ref m);
		}

		/// <summary>
		/// 非表示でウインドウを作成
		/// </summary>
		protected override CreateParams CreateParams
		{
			[System.Security.Permissions.SecurityPermission(System.Security.Permissions.SecurityAction.LinkDemand, Flags = System.Security.Permissions.SecurityPermissionFlag.UnmanagedCode)]
			get
			{
				const int WS_EX_TOOLWINDOW = 0x80;
				const long WS_POPUP = 0x80000000L;
				const int WS_VISIBLE = 0x10000000;
				const int WS_SYSMENU = 0x80000;
				const int WS_MAXIMIZEBOX = 0x10000;

				CreateParams cp = base.CreateParams;
				cp.ExStyle = WS_EX_TOOLWINDOW;
				cp.Style = unchecked((int)WS_POPUP) | WS_VISIBLE | WS_SYSMENU | WS_MAXIMIZEBOX;
				cp.Width = 0;
				cp.Height = 0;

				return cp;
			}
		}

		/// <summary>
		/// 起動されているプレーヤーを調べる
		/// </summary>
		/// <returns></returns>
		private SerachedPlayerTypeFlags CheckRunningPlayerApp()
		{
			SerachedPlayerTypeFlags result = SerachedPlayerTypeFlags.None;

			try
			{
				System.Diagnostics.Process[] ps = System.Diagnostics.Process.GetProcesses();

				foreach(System.Diagnostics.Process procs in ps)
				{
					if(procs.ProcessName == "iTunes")
						result |= SerachedPlayerTypeFlags.iTunes;
					if(procs.ProcessName == "winamp")
						result |= SerachedPlayerTypeFlags.Winamp;

					if(result == SerachedPlayerTypeFlags.All)
						break;
				}
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}

			return result;
		}

		/// <summary>
		/// フォームが開く際に呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void FormMain_Load(object sender, EventArgs e)
		{
			// 設定の読み込み
			try
			{
				_appSetting = AppSetting.Deserialize(Generic.AppSettingFilename);
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
				_appSetting = new AppSetting();
			}

			try
			{
				_skinSetting = Skin.SkinSetting.Deserialize(Generic.AppSkinSettingFilepath);
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
				_skinSetting = new Skin.SkinSetting();
			}

			// フックの開始
			if(Win32SDK.LLKeyboardHook.IsValidHook)
			{
				Win32SDK.LLKeyboardHook.OnKeyDown += _LLKeyboardHook_OnKeyDown;
				Win32SDK.LLKeyboardHook.OnSysKeyDown += _LLKeyboardHook_OnSysKeyDown;
				Win32SDK.LLKeyboardHook.OnKeyUp += _LLKeyboardHook_OnKeyUp;
				Win32SDK.LLKeyboardHook.OnSysKeyUp += _LLKeyboardHook_OnSysKeyUp;
			}

			// コマンドラインでプラグインを強制する場合
			string[] cmds = System.Environment.GetCommandLineArgs();
			if(cmds.Length > 1)
			{
				if(string.Compare(cmds[1], "/winamp", true) == 0)
				{
					_forcedPlayerType = PlayerType.Winamp;
					_appSetting.PlayerType = PlayerType.Winamp;

					_contextMenuStrip_NotifyIcon.Items.Remove(toolStripSeparator3);
					_contextMenuStrip_NotifyIcon.Items.Remove(useiTunesToolStripMenuItem);
					_contextMenuStrip_NotifyIcon.Items.Remove(useWinampToolStripMenuItem);
				}
				else if(string.Compare(cmds[1], "/itunes", true) == 0)
				{
					_forcedPlayerType = PlayerType.iTunes;
					_appSetting.PlayerType = PlayerType.iTunes;

					_contextMenuStrip_NotifyIcon.Items.Remove(toolStripSeparator3);
					_contextMenuStrip_NotifyIcon.Items.Remove(useiTunesToolStripMenuItem);
					_contextMenuStrip_NotifyIcon.Items.Remove(useWinampToolStripMenuItem);
				}
				else
					_forcedPlayerType = PlayerType.Unknwon;
			}
			else
			{
				_forcedPlayerType = PlayerType.Unknwon;
			}

			// 起動時に通知を無効化する場合
			if(_appSetting.InitDisableNotify)
				_enableChangeTrackNotify = false;
			else
				_enableChangeTrackNotify = true;

			// アプリケーションモニタープラグインの準備
			if(!AppMonitorPluginManager.ReloadPlugin(_appSetting.PlayerType, true))
			{
				MessageBox.Show("アプリケーションモニタープラグインのロードに失敗しました。\n使用プレーヤーを Winamp に変更してロードします。\n（＊ iTunesをインストールしていない場合は必ず失敗します。）", Generic.AppTitle, MessageBoxButtons.OK, MessageBoxIcon.Information);

				if(!AppMonitorPluginManager.ReloadPlugin(PlayerType.Winamp, true))
				{
					MessageBox.Show("Winamp モニタープラグインのロードに失敗しました。なう、ぷれ～いんぐは起動できませんでした。（ファイル構成の異常が考えられます。）", Generic.AppTitle, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

					Application.Exit();
					return;
				}

				_appSetting.PlayerType = PlayerType.Winamp;
			}

			// 通知アイコンの初期化
			UpdateNotifyIconImage();
		}

		/// <summary>
		/// フォームが閉じるときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
		{
			// タイマーの全停止
			try
			{
				_timer_PluginEvent.Stop();
				_timer_Intro.Stop();
				_timer_AppSearch.Stop();

				AppMonitorPluginManager.Dispose();
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}


			/* rev. 2013/04/16
			 * 
			 * 終了時に設定保存が失敗する時があるため、設定ダイアログを OK で閉じたときに設定変更するように変更。
			 * */

			/*
			// 設定の保存
			try
			{
				_appSetting.Serialize(Generic.AppSettingFilename);
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}

			try
			{
				_skinSetting.Serialize(Generic.AppSkinSettingFilepath);
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
			 * */
		}

		#region /* キーフック */

		/// <summary>
		/// 低レベルキーフック　キーダウンイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _LLKeyboardHook_OnKeyDown(object sender, Win32SDK.LLKeyboardHook.KeybordHookEventArgs e)
		{
			if(!_appSetting.EnableShortcutKey)
				return;

			try
			{
				// 修飾キーを追加する
				Keys VKkeyWithModifierKey = (Keys)e.KeyCode | ModifierKeys;

				if(VKkeyWithModifierKey == _appSetting.PlaySCKey)
				{
					playToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.PrevSCKey)
				{
					prevToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.NextSCKey)
				{
					nextToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.VolumeUpSCKey)
				{
					volUpToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.VolumeDownSCKey)
				{
					volDownToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.SwitchIntroModeSCKey)
				{
					enableIntroModeToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.SwitchNotifySCKey)
				{
					switchNotifyToolStripMenuItem.PerformClick();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.SwitchUsePlayerSCKey)
				{
					SwitchUsePlayer();
					e.Cancel = true;
				}
				else if(VKkeyWithModifierKey == _appSetting.NotiftPlayingTrackInfoSCKey)
				{
					NotiftTrackInfo();
					e.Cancel = true;
				}
			}
			finally
			{
			}
		}
		/// <summary>
		/// 低レベルキーフック　キーアップイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _LLKeyboardHook_OnKeyUp(object sender, Win32SDK.LLKeyboardHook.KeybordHookEventArgs e)
		{
		}
		/// <summary>
		/// 低レベルキーフック　システムキーダウンイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _LLKeyboardHook_OnSysKeyDown(object sender, Win32SDK.LLKeyboardHook.KeybordHookEventArgs e)
		{
		}
		/// <summary>
		/// 低レベルキーフック　システムキーアップイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _LLKeyboardHook_OnSysKeyUp(object sender, Win32SDK.LLKeyboardHook.KeybordHookEventArgs e)
		{
		}

		#endregion

		#region /* メニューハンドラ */

		/// <summary>
		/// メニュー：設定ダイアログ表示
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void configToolStripMenuItem_Click(object sender, EventArgs e)
		{
			_timer_AppSearch.Stop();

			// タスクトレイアイコンメニュー、通知の無効化
			_notifyIcon.ContextMenuStrip = null;

			bool restoreChangeTrackNotify = false;
			if(_enableChangeTrackNotify)
			{
				restoreChangeTrackNotify = true;
				_enableChangeTrackNotify = false;
			}

			_notifyIcon.MouseDoubleClick -= _notifyIcon_MouseDoubleClick;

			// イントロ再生の無効化
			bool restoreIntorPlayingMode = false;
			if(EnableIntroPlayMode)
			{
				restoreIntorPlayingMode = true;
				EnableIntroPlayMode = false;
			}

			UpdateNotifyIconImage();

			// 設定ダイアログ表示
			FormConfig form = new FormConfig();

			PlayerType prevPlayerType = _appSetting.PlayerType;

			form.AppSetting = _appSetting.DeepCopy();
			form.SkinSetting = _skinSetting.DeepCopy();

			if(form.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
			{
				_appSetting = form.AppSetting;
				_skinSetting = form.SkinSetting;

				/* rev. 2013/04/16
				 * 
				 * 終了時に設定保存が失敗する時があるため、設定ダイアログを OK で閉じたときに設定変更するように変更。
				 * */

				// 設定の保存
				try
				{
					_appSetting.Serialize(Generic.AppSettingFilename);
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}

				try
				{
					_skinSetting.Serialize(Generic.AppSkinSettingFilepath);
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}

				// 設定でプレーヤーが変更された？
				if(_appSetting.PlayerType != prevPlayerType)
				{
					// プレーヤが指定されているならばプレーヤに戻す
					if(_forcedPlayerType != PlayerType.Unknwon)
					{
						_appSetting.PlayerType = _forcedPlayerType;
					}
					else
					{
						if(_appSetting.PlayerType == PlayerType.Unknwon)
						{
							// ばかよけ
							_appSetting.PlayerType = prevPlayerType;
						}
						else
						{
							// 変更されたプレーヤを読み込む
							try
							{
								AppMonitorPluginManager.ReloadPlugin(_appSetting.PlayerType, true);
							}
							catch(Exception ex)
							{
								Generic.UpdateLogFile(ex);
							}
						}
					}
				}
			}

			_timer_AppSearch.Start();

			// タスクトレイアイコンメニュー、通知有無の復帰
			_notifyIcon.ContextMenuStrip = _contextMenuStrip_NotifyIcon;
			if(restoreChangeTrackNotify)
				_enableChangeTrackNotify = true;

			_notifyIcon.MouseDoubleClick += _notifyIcon_MouseDoubleClick;

			// イントロ再生設定の復帰
			if(restoreIntorPlayingMode)
				EnableIntroPlayMode = true;

			UpdateNotifyIconImage();
		}
		/// <summary>
		/// メニュー：終了
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void exitToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Close();
		}
		/// <summary>
		/// メニュー：ヘルプ
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void viewHelpToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				System.Diagnostics.Process.Start(_appSetting.HelpURL);
			}
			catch(Exception)
			{
			}
		}
		/// <summary>
		/// メニュー：バージョン表示
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			// タスクトレイアイコンメニュー無効化
			_notifyIcon.ContextMenuStrip = null;

			FormAbout form = new FormAbout();
			form.ShowDialog();

			// タスクトレイアイコンメニュー復帰
			_notifyIcon.ContextMenuStrip = _contextMenuStrip_NotifyIcon;
		}
		/// <summary>
		/// メニュー：通知の有効/無効切り替え （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void switchNotifyToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Action method = () =>
			{
				lock(_syncObj)
				{
					_enableChangeTrackNotify = !_enableChangeTrackNotify;

					UpdateNotifyIconImage();
				}
			};

// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}
		/// <summary>
		/// メニュー：再生 （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void playToolStripMenuItem_Click(object sender, EventArgs e)
		{
			ControlPlayer(PluginBase.PlayerControl.SwitchPlay);
		}
		/// <summary>
		/// メニュー：次の曲 （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void nextToolStripMenuItem_Click(object sender, EventArgs e)
		{
			ControlPlayer(PluginBase.PlayerControl.Next);
		}
		/// <summary>
		/// メニュー：前の曲 （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void prevToolStripMenuItem_Click(object sender, EventArgs e)
		{
			ControlPlayer(PluginBase.PlayerControl.Prev);
		}
		/// <summary>
		/// メニュー：音量UP （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void volUpToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Action method = () =>
			{
				try
				{
					lock(_syncObj)
					{
						int vol = AppMonitorPluginManager.SoundVolume;
						if(vol < 0)
							return;

						vol += 5;

						if(vol > 100)
							vol = 100;

						AppMonitorPluginManager.SoundVolume = vol;
					}
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}
			};

// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}
		/// <summary>
		/// メニュー：音量Down （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void volDownToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Action method = () =>
			{
				try
				{
					lock(_syncObj)
					{
						int vol = AppMonitorPluginManager.SoundVolume;
						if(vol < 0)
							return;

						vol -= 5;

						if(vol < 0)
							vol = 0;

						AppMonitorPluginManager.SoundVolume = vol;
					}
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}
			};

// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}
		/// <summary>
		/// メニュー：イントロ再生設定 （スレッドセーフ。PerformClick で実行しても安全です）
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void enableIntroModeToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Action method = () =>
			{
				try
				{
					lock(_syncObj)
					{
						EnableIntroPlayMode = !EnableIntroPlayMode;

						UpdateNotifyIconImage();
					}
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}
			};

// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}
		/// <summary>
		/// メニュー： iTunes選択
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void useiTunesToolStripMenuItem_Click(object sender, EventArgs e)
		{
			// イントロの強制終了
			EnableIntroPlayMode = false;

			if(AppMonitorPluginManager.SwitchPlugin(PlayerType.iTunes, true))
				_appSetting.PlayerType = PlayerType.iTunes;

			UpdateNotifyIconImage();
		}
		/// <summary>
		/// メニュー：Wianmp選択
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void useWinampToolStripMenuItem_Click(object sender, EventArgs e)
		{
			// イントロの強制終了
			EnableIntroPlayMode = false;

			if(AppMonitorPluginManager.SwitchPlugin(PlayerType.Winamp, true))
				_appSetting.PlayerType = PlayerType.Winamp;

			UpdateNotifyIconImage();
		}
		/// <summary>
		/// メニューが開くときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _contextMenuStrip_NotifyIcon_Opening(object sender, CancelEventArgs e)
		{
			try
			{
				#region // 接続状況でプレーヤーコントロールの有効/無効設定するメソッド setMenuEnable
				Action<bool> setMenuEnable_method = (enable) =>
				{
					prevToolStripMenuItem.Enabled = enable;
					nextToolStripMenuItem.Enabled = enable;
					playToolStripMenuItem.Enabled = enable;
					volUpToolStripMenuItem.Enabled = enable;
					volDownToolStripMenuItem.Enabled = enable;
				};
				#endregion

				if(AppMonitorPluginManager.IsValid)
				{
					setMenuEnable_method(AppMonitorPluginManager.IsConnected);

					if(AppMonitorPluginManager.IsPlaying)
					{
						playToolStripMenuItem.Image = NauPlaying.Properties.Resources.MenuPause;
						playToolStripMenuItem.Text = "一次停止(&S)";
					}
					else
					{
						playToolStripMenuItem.Image = NauPlaying.Properties.Resources.MenuPlay;
						playToolStripMenuItem.Text = "再生(&P)";
					}
				}
				else
					setMenuEnable_method(false);


				// プレーヤタイプでチェックの変更
				if(_appSetting.PlayerType == PlayerType.iTunes)
				{
					useiTunesToolStripMenuItem.Checked = true;
					useWinampToolStripMenuItem.Checked = false;
				}
				else
				{
					useiTunesToolStripMenuItem.Checked = false;
					useWinampToolStripMenuItem.Checked = true;
				}

				// イントロタイマー動作中はプレーヤチェンジの禁止
				if(_timer_Intro.Enabled)
				{
					useiTunesToolStripMenuItem.Enabled = false;
					useWinampToolStripMenuItem.Enabled = false;

					useiTunesToolStripMenuItem.ToolTipText = "イントロ動作中はプレーヤ切り替えはできません。";
					useWinampToolStripMenuItem.ToolTipText = "イントロ動作中はプレーヤ切り替えはできません。";
				}
				else
				{
					useiTunesToolStripMenuItem.Enabled = true;
					useWinampToolStripMenuItem.Enabled = true;

					// ショートカットキーが有効な場合はツールチップ表示セット
					if(_appSetting.EnableShortcutKey)
					{
						useiTunesToolStripMenuItem.ToolTipText = "プレーヤ切り替え\nｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.SwitchUsePlayerSCKeyAsString;
						useWinampToolStripMenuItem.ToolTipText = "プレーヤ切り替え\nｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.SwitchUsePlayerSCKeyAsString;
					}
					else
					{
						useiTunesToolStripMenuItem.ToolTipText = "";
						useWinampToolStripMenuItem.ToolTipText = "";
					}
				}

				// ショートカットキーが有効な場合はツールチップ表示セット
				if(_appSetting.EnableShortcutKey)
				{
					prevToolStripMenuItem.ToolTipText = "ｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.PrevSCKeyAsString;
					nextToolStripMenuItem.ToolTipText = "ｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.NextSCKeyAsString;
					playToolStripMenuItem.ToolTipText = "ｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.PlaySCKeyAsString;
					volUpToolStripMenuItem.ToolTipText = "ｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.VolumeUpSCKeyAsString;
					volDownToolStripMenuItem.ToolTipText = "ｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.VolumeDownSCKeyAsString;

					switchNotifyToolStripMenuItem.ToolTipText = "通知のON/OFFを切り替えます。\nｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.SwitchNotifySCKeyAsString;
					enableIntroModeToolStripMenuItem.ToolTipText = "イントロモードのON/OFFを切り替えます。\nｼｮｰﾄｶｯﾄｷｰ : " + _appSetting.SwitchIntroModeSCKeyAsString;
				}
				else
				{
					prevToolStripMenuItem.ToolTipText = "";
					nextToolStripMenuItem.ToolTipText = "";
					playToolStripMenuItem.ToolTipText = "";
					volUpToolStripMenuItem.ToolTipText = "";
					volDownToolStripMenuItem.ToolTipText = "";

					switchNotifyToolStripMenuItem.ToolTipText = "通知のON/OFFを切り替えます。";
					enableIntroModeToolStripMenuItem.ToolTipText = "イントロモードのON/OFFを切り替えます。";
				}

				// 通知の有無
				switchNotifyToolStripMenuItem.Checked = _enableChangeTrackNotify;
				// イントロモードの場合
				enableIntroModeToolStripMenuItem.Checked = EnableIntroPlayMode;
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
		}

		#endregion

		/// <summary>
		/// タスクトレイアイコンのダブルクリック
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _notifyIcon_MouseDoubleClick(object sender, MouseEventArgs e)
		{
			if(e.Button == System.Windows.Forms.MouseButtons.Left)
			{
				_enableChangeTrackNotify = !_enableChangeTrackNotify;

				UpdateNotifyIconImage();
			}
		}

		/// <summary>
		/// アニメーション開始イベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _formBoard_OnBeginAnimationEvent(object sender, EventArgs e)
		{
		}

		/// <summary>
		/// アニメーション終了イベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _formBoard_OnEndAnimationEvent(object sender, EventArgs e)
		{
			try
			{
				if(_formBoard != null)
				{
					if(_formBoard.IsAnimation)
						_formBoard.StopAnimation();

					_formBoard.Close();
					_formBoard.Dispose();
					_formBoard = null;
				}
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
		}

		/// <summary>
		/// 現在の状態に合わせて通知アイコンを設定
		/// </summary>
		private void UpdateNotifyIconImage()
		{
			try
			{
				if(AppMonitorPluginManager.IsConnected)
				{
					if(_enableChangeTrackNotify)
					{
						if(EnableIntroPlayMode)
							_notifyIcon.Icon = NauPlaying.Properties.Resources.NotifiIcon_Intro;
						else
							_notifyIcon.Icon = NauPlaying.Properties.Resources.NotifiIcon;
					}
					else
					{
						if(EnableIntroPlayMode)
							_notifyIcon.Icon = NauPlaying.Properties.Resources.NotifiIcon_Disable_Intro;
						else
							_notifyIcon.Icon = NauPlaying.Properties.Resources.NotifiIcon_Disable;
					}
				}
				else
				{
					if(_enableChangeTrackNotify)
						_notifyIcon.Icon = NauPlaying.Properties.Resources.NotifiIcon_Disconnect;
					else
						_notifyIcon.Icon = NauPlaying.Properties.Resources.NotifiIcon_Disable_Disconnect;
				}
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
		}

		/// <summary>
		/// 現在再生中の曲の通知（スレッドセーフ）
		/// </summary>
		private void NotiftTrackInfo()
		{
			#region /* デフォルトのイメージを読み込みメソッド (例外対策つき) */
			Func<Bitmap> readDefaultImage_method = () =>
			{
				// null で透過
				Bitmap bmp = null;

				if(_appSetting.DispNoneArtworkImage)
				{
					try
					{
						using(FileStream fs = new FileStream(Generic.AppPath + "\\" + _appSetting.NoneArtworkImageFilename, System.IO.FileMode.Open, System.IO.FileAccess.Read))
						{
							bmp = new Bitmap(fs);
						}
					}
					catch(Exception ex)
					{
						Generic.UpdateLogFile(ex);
						return null;
					}
				}

				return bmp;
			};
			#endregion

			Action method = () =>
			{
				lock(_syncObj)
				{
					if(!_enableChangeTrackNotify)
						return;

					// トラック情報を取得
					PluginBase.TrackInfo track = AppMonitorPluginManager.CurrentTrackInfo;
					if(track == null)
						return;

					PluginBase.TrackInfo trackInfo = track.DeepCopy();

					// アートワークの取得
					Bitmap artwork = null;
					try
					{
						// アートワークが存在する場合には読み込みを行う
						if(trackInfo.ArtworkFilepath.Length > 0)
						{
							using(FileStream fs = new FileStream(trackInfo.ArtworkFilepath, System.IO.FileMode.Open, System.IO.FileAccess.Read))
							{
								artwork = new Bitmap(Image.FromStream(fs));
							}
						}
						else
						{
							artwork = readDefaultImage_method();

							// ファイルパスは有効？
							if(trackInfo.FilePath.Length > 0)
							{
								try
								{
									NTag.TagReader reader = new NTag.TagReader(trackInfo.FilePath);

									if(reader.IsValidFile())
									{
										reader.Read();
										NTag.Tag tag = reader.Tag;

										if(tag.Artwork != null)
											artwork = (Bitmap)tag.Artwork;
									}
								}
								catch
								{
								}
							}
						}
					}
					catch(Exception ex)
					{
						artwork = readDefaultImage_method();
						Generic.UpdateLogFile(ex);
					}

					// ボードウインドウの生成
					try
					{
						// 表示中の場合は終了
						if(_formBoard != null)
						{
							if(_formBoard.IsAnimation)
								_formBoard.StopAnimation();

							// _formBoard.StopAnimation メソッドでアニメーション終了イベントが出ないのでここで後始末をする
							_formBoard.Close();
							_formBoard.Dispose();
							_formBoard = null;
						}

						_formBoard = new FormBoard();

						_formBoard.OnBeginAnimationEvent += _formBoard_OnBeginAnimationEvent;
						_formBoard.OnEndAnimationEvent += _formBoard_OnEndAnimationEvent;

						_formBoard.InitializeBoard(_appSetting, _skinSetting, new BoardData(trackInfo.Format(_appSetting.TextFormat), artwork));

						_formBoard.ShowWithoutActiveTopMost(true);
						_formBoard.StartAnimation();
					}
					catch(Exception ex)
					{
						Generic.UpdateLogFile(ex);
					}
				}
			};

// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}

		/// <summary>
		/// 使用プレーヤを切り替える（スレッドセーフ）
		/// </summary>
		private void SwitchUsePlayer()
		{
			// プレーヤが指定されているならば何もしない
			if(_forcedPlayerType != PlayerType.Unknwon)
				return;
			// イントロ動作中はプレーヤ切り替え無効
			if(_timer_Intro.Enabled)
				return;

			Action method = () =>
			{
				lock(_syncObj)
				{
					if(AppMonitorPluginManager.PlayerType == PlayerType.iTunes)
					{
						if(AppMonitorPluginManager.SwitchPlugin(PlayerType.Winamp, true))
						{
							_appSetting.PlayerType = PlayerType.Winamp;

							if(_appSetting.EnableBalloonTip)
							{
								_notifyIcon.BalloonTipText = "Wianmpに切り替えたよ！";
								_notifyIcon.ShowBalloonTip(_appSetting.BalloonTipTimeout);
							}
						}
					}
					else
					{
						if(AppMonitorPluginManager.SwitchPlugin(PlayerType.iTunes, true))
						{
							_appSetting.PlayerType = PlayerType.iTunes;

							if(_appSetting.EnableBalloonTip)
							{
								_notifyIcon.BalloonTipText = "iTunesに切り替えたよ！";
								_notifyIcon.ShowBalloonTip(_appSetting.BalloonTipTimeout);
							}
						}
					}

					UpdateNotifyIconImage();
				}
			};

// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}

		/// <summary>
		/// プラグインイベント監視タイマー
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _timer_Tick(object sender, EventArgs e)
		{
			try
			{
				lock(_pluginEventTimer_lockObj)
				{
					// アプリケーションモニターのイベントをチェック
					PluginBase.AppMonitorEvent appEvent;
					while(AppMonitorPluginManager.PopAppMonitorEvent(out appEvent))
					{
						switch(appEvent)
						{
						case PluginBase.AppMonitorEvent.ChangePlayingTrack:
							NotiftTrackInfo();
							break;

						case PluginBase.AppMonitorEvent.StopPlayingTrack:
							break;

						case PluginBase.AppMonitorEvent.Connect:
							UpdateNotifyIconImage();

							// 現在の設定ボリュームを保存
							int vol = AppMonitorPluginManager.SoundVolume;
							if(vol >= 0)
								_introFadeUserVolume = vol;
							break;

						case PluginBase.AppMonitorEvent.Disconnect:
							UpdateNotifyIconImage();
							AppMonitorPluginManager.ReloadPlugin(_appSetting.PlayerType, false);
							break;

						case PluginBase.AppMonitorEvent.PluginErrorEvent:
							if(AppMonitorPluginManager.LastException != null)
								Generic.UpdateLogFile(AppMonitorPluginManager.LastException);
							break;
						}
					}
				}
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
		}

		/// <summary>
		/// イントロ制御タイマー
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _timer_Intro_Tick(object sender, EventArgs e)
		{
			if(!AppMonitorPluginManager.IsValid)
				return;

			try
			{
				lock(_introTimer_lockObj)
				{
					// イントロ設定で次の曲へ送るとともに、フェードアウト設定
					if(AppMonitorPluginManager.PlayerPosition >= _intorChangeNextTime && !IsIntroFeading)
					{
						UpdateIntroVolumeSetting();

						if(_appSetting.EnableIntroFadeout)
							_introFadeoutFlag = true;
						else
							ControlPlayer(PluginBase.PlayerControl.Next);
					}

					// フェードイン動作
					if(_introFadeinFlag)
					{
						_introFadeCurrentVolume += _introFadeVolumeDelta;
						if(_introFadeCurrentVolume >= _introFadeUserVolume)
							_introFadeCurrentVolume = _introFadeUserVolume;

						AppMonitorPluginManager.SoundVolume = _introFadeCurrentVolume;

						if(_introFadeCurrentVolume >= _introFadeUserVolume)
							_introFadeinFlag = false;
					}
					// フェードアウト動作
					else if(_introFadeoutFlag)
					{
						_introFadeCurrentVolume -= _introFadeVolumeDelta;
						if(_introFadeCurrentVolume <= 0)
							_introFadeCurrentVolume = 0;

						AppMonitorPluginManager.SoundVolume = _introFadeCurrentVolume;

						if(_introFadeCurrentVolume <= 0)
						{
							_introFadeoutFlag = false;
							ControlPlayer(PluginBase.PlayerControl.Next);

							if(!_appSetting.EnableIntroFadein)
								AppMonitorPluginManager.SoundVolume = _introFadeUserVolume;
						}
					}
				}
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
		}

		/// <summary>
		/// プレーヤサーチタイマー
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _timer_AppSearch_Tick(object sender, EventArgs e)
		{
			try
			{
				lock(_appSearchTimer_lockObj)
				{
					if(!_appSetting.EnablePlayerSearch)
						return;

					SerachedPlayerTypeFlags players = CheckRunningPlayerApp();

					// 全プレーヤーが起動している状態で
					if(players == SerachedPlayerTypeFlags.All)
					{
						// 強制切り替えならば
						if(_appSetting.EnableForcedChangePlayer)
						{
							switch(_appSetting.PlayerSearchPriority)
							{
							// iTunes 優先
							case PlayerSearchPriority.iTunes:
								if(_appSetting.PlayerType != PlayerType.iTunes)
									SwitchUsePlayer();
								break;

							// Winamp 優先
							case PlayerSearchPriority.Winamp:
								if(_appSetting.PlayerType != PlayerType.Winamp)
									SwitchUsePlayer();
								break;
							}
						}
					}
					else
					{
						// iTunes のみが起動している状態で
						if((players & SerachedPlayerTypeFlags.iTunes) == SerachedPlayerTypeFlags.iTunes)
						{
							if(_appSetting.PlayerType != PlayerType.iTunes)
								SwitchUsePlayer();
						}
						// Winamp のみが起動している状態で
						if((players & SerachedPlayerTypeFlags.Winamp) == SerachedPlayerTypeFlags.Winamp)
						{
							if(_appSetting.PlayerType != PlayerType.Winamp)
								SwitchUsePlayer();
						}
					}
				}
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
			}
		}

		/// <summary>
		/// // 現在のイントロ関係に使用するボリュームを更新
		/// </summary>
		/// <returns></returns>
		private bool UpdateIntroVolumeSetting()
		{
			try
			{
				int vol = AppMonitorPluginManager.SoundVolume;
				if(vol < 0)
					return false;

				_introFadeUserVolume = vol;

				// ボリューム変化量の計算
				if(_appSetting.IntroFadeTime <= 0)
					_introFadeVolumeDelta = 1;
				else if(_appSetting.IntroFadeTime > 10000)
					_introFadeVolumeDelta = _introFadeUserVolume / 100;
				else
					_introFadeVolumeDelta = _introFadeUserVolume / (_appSetting.IntroFadeTime / 100);

				if(_introFadeVolumeDelta <= 0)
					_introFadeVolumeDelta = 1;
				else if(_introFadeVolumeDelta > 100)
					_introFadeVolumeDelta = 100;
			}
			catch(Exception ex)
			{
				Generic.UpdateLogFile(ex);
				return false;
			}

			return true;
		}

		/// <summary>
		/// イントロモードのフェード設定を更新
		/// </summary>
		/// <param name="seekCurrentPostion">true で現在の再生位置から次回曲送り時間を設定</param>
		/// <returns></returns>
		private bool UpdateIntroFeadSettting(bool seekCurrentPostion)
		{
			#region /* イントロ設定を無効化（通常再生を行う）するメソッド */
			Action disableIntoroSetting_method = () =>
			{
				_intorChangeNextTime = int.MaxValue;

				_introFadeinFlag = false;
				_introFadeoutFlag = false;
			};
			#endregion

			if(!AppMonitorPluginManager.IsValid)
				return false;
			if(!EnableIntroPlayMode)
				return false;

			try
			{
				int introPlayTime = 0;
				int introStartTime = 0;

				_introFadeCurrentVolume = _introFadeUserVolume;

				// 曲送り時間の設定
				PluginBase.TrackInfo currentTrack = AppMonitorPluginManager.CurrentTrackInfo;
				if(currentTrack != null)
				{
					// 単位指定に合わせて再生時間を設定
					if(_appSetting.IntroTimeUnit == IntroTimeUnit.Sec)
					{
						introPlayTime = _appSetting.IntroPlayTime;
						introStartTime = _appSetting.IntroStartTime;
					}
					else
					{
						introPlayTime = (int)((float)currentTrack.Duration * (float)_appSetting.IntroPlayTime / 100.0f);
						introStartTime = (int)((float)currentTrack.Duration * (float)_appSetting.IntroStartTime / 100.0f);
					}

					// フェードインを考慮した総再生時間[s]を算出
					int totalTime = 0;
					// 現在の再生位置から？
					if(seekCurrentPostion)
					{
						int pos = AppMonitorPluginManager.PlayerPosition;
						if(pos < 0)
							pos = 0;

						totalTime = pos + introPlayTime;

						if(_appSetting.EnableIntroFadeout)
							totalTime += _appSetting.IntroFadeTime / 1000;
					}
					// 曲の頭？
					else
					{
						totalTime = introStartTime + introPlayTime;

						if(_appSetting.EnableIntroFadein)
							totalTime += _appSetting.IntroFadeTime / 1000;
						if(_appSetting.EnableIntroFadeout)
							totalTime += _appSetting.IntroFadeTime / 1000;
					}

					// 再生中の総再生時間より小さい場合イントロモードの設定を行う
					if(totalTime < currentTrack.Duration)
					{
						// 次の曲のスキップ時間
						_intorChangeNextTime = totalTime;

						// 頭からの場合、フェードイン設定、再生開始位置へシークを行う
						if(!seekCurrentPostion)
						{
							if(_appSetting.EnableIntroFadein)
							{
								AppMonitorPluginManager.SoundVolume = 0;
								_introFadeCurrentVolume = 0;

								_introFadeinFlag = true;
							}

							if(_appSetting.IntroStartTime > 0)
								AppMonitorPluginManager.PlayerPosition = introStartTime;
						}
					}
					else
					{
						disableIntoroSetting_method();
					}
				}
				else
				{
					disableIntoroSetting_method();
				}
			}
			catch(Exception ex)
			{
				disableIntoroSetting_method();

				Generic.UpdateLogFile(ex);
				return false;
			}

			return true;
		}

		/// <summary>
		/// プレーヤを操作する
		/// （イントロモードが有効な場合はフェード設定も行う）
		/// （スレッドセーフ）
		/// </summary>
		/// <param name="type"></param>
		private void ControlPlayer(PluginBase.PlayerControl type)
		{
			if(!AppMonitorPluginManager.IsValid)
				return;

			Action method = () =>
			{
				try
				{
					lock(_syncObj)
					{
						switch(type)
						{
						case PluginBase.PlayerControl.Play:
							AppMonitorPluginManager.ControlPlayer(PluginBase.PlayerControl.Play);
							break;
						case PluginBase.PlayerControl.Stop:
							AppMonitorPluginManager.ControlPlayer(PluginBase.PlayerControl.Stop);
							break;
						case PluginBase.PlayerControl.SwitchPlay:
							AppMonitorPluginManager.ControlPlayer(PluginBase.PlayerControl.SwitchPlay);
							break;

						case PluginBase.PlayerControl.Next:
							// イントロモードのフェード動作は曲変更を無視
							if(!IsIntroFeading)
							{
								AppMonitorPluginManager.ControlPlayer(PluginBase.PlayerControl.Next);

								if(EnableIntroPlayMode)
									UpdateIntroFeadSettting(false);
							}
							break;

						case PluginBase.PlayerControl.Prev:
							// イントロモードのフェード動作は曲変更を無視
							if(!IsIntroFeading)
							{
								AppMonitorPluginManager.ControlPlayer(PluginBase.PlayerControl.Prev);

								if(EnableIntroPlayMode)
									UpdateIntroFeadSettting(false);
							}
							break;
						}
					}
				}
				catch(Exception ex)
				{
					Generic.UpdateLogFile(ex);
				}
			};

			// iTunes の場合、直に呼んだり、Invokeでは COM の例外が起きるので、非同期呼び出し（改善が必要か?）
			BeginInvoke(method);
		}

		
	}
}
