﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using Win32SDK;

namespace NauPlaying
{
	/// <summary>
	/// 領域編集フォーム
	/// </summary>
	public partial class FormRegionEditor : Form
	{
		/// <summary>
		/// 選択している領域識別のための列挙子
		/// </summary>
		private enum SelectRgnType
		{
			/// <summary>
			/// 選択されていいない
			/// </summary>
			None,

			/// <summary>
			/// アートワークが選択されている
			/// </summary>
			Artwork,
			/// <summary>
			/// テキストが選択されている
			/// </summary>
			Text,
			/// <summary>
			/// バックグランドイメージが選択されている
			/// </summary>
			BkImage,
		}

		/// <summary>
		/// マウスダウンイベント検地時のマウス座標
		/// </summary>
		private Point _mouseButtonDownPoint;
		
		/// <summary>
		/// マウスダウンイベント検地時のヒットテスト結果
		/// </summary>
		private RegionalType _hitTestResult = RegionalType.None;
		
		/// <summary>
		/// ピクチャボックス用のベースイメージ
		/// </summary>
		private Bitmap _baseImage;
		/// <summary>
		/// スクリーンのイメージ
		/// </summary>
		private Bitmap _screenImage;

		/// <summary>
		/// 四角形領域マップ
		/// </summary>
		private Dictionary<SelectRgnType, RegionRectEditor> _regionRectMap;

		/// <summary>
		/// ステータスバーの高さを取得
		/// </summary>
		private int StatusHeight
		{
			get { return _statusStrip.Size.Height; }
		}
		/// <summary>
		/// メニューの高さを取得
		/// </summary>
		private int MenuHeight
		{
			get { return _menuStrip.Size.Height; }
		}

		/// <summary>
		/// 現在選択されたいる領域の取得、設定
		/// </summary>
		private SelectRgnType SelectedRgnType
		{
			get
			{
				if(artworkToolStripMenuItem.Checked)
					return SelectRgnType.Artwork;
				if(textToolStripMenuItem.Checked)
					return SelectRgnType.Text;
				if(bkImageToolStripMenuItem.Checked)
					return SelectRgnType.BkImage;

				return SelectRgnType.None;
			}
			set
			{
				switch(value)
				{
				case SelectRgnType.Artwork:
					artworkToolStripMenuItem.Checked = true;
					textToolStripMenuItem.Checked = false;
					bkImageToolStripMenuItem.Checked = false;
					break;

				case SelectRgnType.Text:
					artworkToolStripMenuItem.Checked = false;
					textToolStripMenuItem.Checked = true;
					bkImageToolStripMenuItem.Checked = false;
					break;

				case SelectRgnType.BkImage:
					artworkToolStripMenuItem.Checked = false;
					textToolStripMenuItem.Checked = false;
					bkImageToolStripMenuItem.Checked = true;
					break;

				default:
					break;
				}
			}
		}

		/// <summary>
		/// 頂点指示円の半径
		/// </summary>
		private int _vertexCircleR;
		/// <summary>
		/// 点や線がヒットしたと判定するマージン
		/// </summary>
		private int _hitMargin;
		/// <summary>
		/// 四隅の点の描画色
		/// </summary>
		private Color _vertexCircleFillColor;
		/// <summary>
		/// ライン描画の際のライン色
		/// </summary>
		private Color _linePenColor;

		/// <summary>
		/// ボードサイズの取得、設定
		/// （表示前に必ず設定すること）
		/// </summary>
		public Size BoardSize { set; get; }
		/// <summary>
		/// アートワーク表示領域の取得、設定
		/// （表示前に必ず設定すること）
		/// </summary>
		public Rectangle ArtworkDispRect { set; get; }
		/// <summary>
		/// テキスト表示領域の取得、設定
		/// （表示前に必ず設定すること）
		/// </summary>
		public Rectangle TextDispRect { set; get; }
		/// <summary>
		/// バックグランドイメージ表示領域の取得、設定
		/// （表示前に必ず設定すること）
		/// </summary>
		public Rectangle BkImageDispRect { set; get; }
		/// <summary>
		/// バックグランドイメージの取得、設定
		/// イメージなしの場合は null を設定。
		/// （表示前に必ず設定すること）
		/// </summary>
		public Bitmap BkImage { set; get; }

		/// <summary>
		/// アートワーク領域色
		/// </summary>
		public Color ArtworkColor { set; get; }
		/// <summary>
		/// テキスト領域色
		/// </summary>
		public Color TextColor { set; get; }
		/// <summary>
		/// バックグランド領域色
		/// </summary>
		public Color BkImageColor { set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		public FormRegionEditor()
		{
			_vertexCircleR = 3;
			_hitMargin = 3;
			_vertexCircleFillColor = Color.FromArgb(255, 0, 0);
			_linePenColor = Color.FromArgb(0, 0, 0);

			BoardSize = new System.Drawing.Size(512, 96);
			ArtworkDispRect = new Rectangle(0, 0, 96, 96);
			TextDispRect = new Rectangle(96, 0, 320, 96);
			BkImageDispRect = new Rectangle(416, 0, 96, 96);

			BkImage = Skin.SkinRender.CreateNoneArtworkImage(BkImageDispRect.Size);

			ArtworkColor = Color.FromArgb(96, 255, 200, 150);
			TextColor = Color.FromArgb(96, 255, 150, 200);
			BkImageColor = Color.FromArgb(96, 200, 255, 200);

			// コントロールの初期化
			InitializeComponent();
		}

		/// <summary>
		/// フォーム表示前イベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void FormRegionSelect_Load(object sender, EventArgs e)
		{
			// 領域編集のためのマップ生成と初期化
			_regionRectMap = new Dictionary<SelectRgnType, RegionRectEditor>();
			_regionRectMap.Add(SelectRgnType.Artwork, new RegionRectEditor(BoardSize));
			_regionRectMap.Add(SelectRgnType.Text, new RegionRectEditor(BoardSize));
			_regionRectMap.Add(SelectRgnType.BkImage, new RegionRectEditor(BoardSize));

			_regionRectMap[SelectRgnType.Artwork].RegionRect = ArtworkDispRect;
			_regionRectMap[SelectRgnType.Artwork].RgnFillColor = ArtworkColor;
			_regionRectMap[SelectRgnType.Artwork].VertexCircleR= _vertexCircleR;
			_regionRectMap[SelectRgnType.Artwork].VertexCircleFillColor = _vertexCircleFillColor;
			_regionRectMap[SelectRgnType.Artwork].HitMargin = _hitMargin;
			_regionRectMap[SelectRgnType.Artwork].LinePenColor = _linePenColor;
			_regionRectMap[SelectRgnType.Artwork].IsSquareMode = true;
			_regionRectMap[SelectRgnType.Artwork].Text = "ｱｰﾄﾜｰｸ";
			
			_regionRectMap[SelectRgnType.Text].RegionRect = TextDispRect;
			_regionRectMap[SelectRgnType.Text].RgnFillColor = TextColor;
			_regionRectMap[SelectRgnType.Text].VertexCircleR = _vertexCircleR;
			_regionRectMap[SelectRgnType.Text].VertexCircleFillColor = _vertexCircleFillColor;
			_regionRectMap[SelectRgnType.Text].HitMargin = _hitMargin;
			_regionRectMap[SelectRgnType.Text].LinePenColor = _linePenColor;
			_regionRectMap[SelectRgnType.Text].Text = "ﾃｷｽﾄ";

			_regionRectMap[SelectRgnType.BkImage].RegionRect = BkImageDispRect;
			_regionRectMap[SelectRgnType.BkImage].RgnFillColor = BkImageColor;
			_regionRectMap[SelectRgnType.BkImage].VertexCircleR = _vertexCircleR;
			_regionRectMap[SelectRgnType.BkImage].VertexCircleFillColor = _vertexCircleFillColor;
			_regionRectMap[SelectRgnType.BkImage].HitMargin = _hitMargin;
			_regionRectMap[SelectRgnType.BkImage].LinePenColor = _linePenColor;
			_regionRectMap[SelectRgnType.BkImage].Text = "ﾊﾞｯｸｸﾞﾗﾝﾄﾞ\nｲﾒｰｼﾞ";

			// フォームの大きさ調整
			if(BkImage != null)
			{
				Size bk = BkImage.Size;
				if(bk != BoardSize)
				{
					if(MessageBox.Show("バックグランドイメージのサイズと設定されているボードサイズが異なります。ボードサイズをバックグランドイメージのサイズに変更しますか？", Generic.AppTitle, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == System.Windows.Forms.DialogResult.Yes)
					{
						BoardSize = bk;
					}
				}
			}

			_pictureBox.Location = new Point(0, MenuHeight);
			_pictureBox.Size = BoardSize;

			ClientSize = new Size(BoardSize.Width, BoardSize.Height + MenuHeight + StatusHeight);

			// 領域の描画
			SelectedRgnType = SelectRgnType.Artwork;
			CaptureScreenImage();
			UpdatePictureBox();

			RegionRectEditor rrs = _regionRectMap[SelectedRgnType];
			string str = string.Format("(X={0}, Y={1}, W={2}, H={3})", rrs.RegionRect.Left, rrs.RegionRect.Top, rrs.RegionRect.Width, rrs.RegionRect.Height);
			_toolStripStatusLabel_Msg.Text = str;

			// イメージなしでメニューを無効化
			if(BkImage == null)
				bkImageToolStripMenuItem.Enabled = false;
		}

		/// <summary>
		/// ピクチャーボックスの表示更新
		/// </summary>
		private void UpdatePictureBox()
		{
			if(SelectedRgnType == SelectRgnType.None)
				return;

			RegionRectEditor selectedRRS = _regionRectMap[SelectedRgnType];
			selectedRRS.UpdateRegionRect();

			if(_pictureBox.Image != null)
				_pictureBox.Image.Dispose();

			if(_baseImage != null)
				_baseImage.Dispose();

			_baseImage = new Bitmap(BoardSize.Width, BoardSize.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

			// 描画処理
			using(Graphics g = Graphics.FromImage(_baseImage))
			{
				// 領域を描画するメソッド
				Action<SelectRgnType, Color> DrawRegion = (type, col) =>
				{
					using(Pen pen = new Pen(col, 1))
					using(Brush brush = new SolidBrush(col))
					{
						RegionRectEditor rrs = _regionRectMap[type];
						if(!rrs.IsMoving && type != SelectedRgnType)
						{
							g.FillRectangle(brush, rrs.RegionRect);
							g.DrawRectangle(pen, rrs.RegionRect);
						}
					}
				};

				// 背景色で塗りつぶし
				Rectangle rect = RectangleToScreen(new Rectangle(_pictureBox.Location, _pictureBox.Size));
				g.DrawImage(_screenImage, new Rectangle(0, 0, BoardSize.Width, BoardSize.Height), rect, GraphicsUnit.Pixel);

				// バックグランドイメージ領域の描画
				if(BkImage != null)
				{
					// イメージ描画
					RegionRectEditor rrs = _regionRectMap[SelectRgnType.BkImage];

					Rectangle src = new Rectangle(0, 0, BkImage.Width, BkImage.Height);
					Rectangle dst;

					if(rrs.IsMoving)
						dst = _regionRectMap[SelectRgnType.BkImage].MovingRegionRect;
					else
						dst = _regionRectMap[SelectRgnType.BkImage].RegionRect;

					g.DrawImage(BkImage, dst, src, GraphicsUnit.Pixel);

					// 領域描画
					DrawRegion(SelectRgnType.BkImage, BkImageColor);
				}

				// アートワーク領域の描画
				DrawRegion(SelectRgnType.Artwork, ArtworkColor);

				// テキスト領域の描画
				DrawRegion(SelectRgnType.Text, TextColor);

				// 編集中領域の描画
				g.DrawImage(selectedRRS.RegionRectBtmap, new Point());
			}

			_pictureBox.Image = _baseImage;
		}

		/// <summary>
		/// キーのトラップ（リターンキー、ESCキーでの終了サポート）
		/// </summary>
		/// <param name="keyData"></param>
		/// <returns></returns>
		protected override bool ProcessDialogKey(Keys keyData)
		{
			switch(keyData)
			{
			case Keys.Escape:
				DialogResult = System.Windows.Forms.DialogResult.Cancel;
				Close();
				return false;

			case Keys.Return:
				DialogResult = System.Windows.Forms.DialogResult.OK;
				Close();
				return false;
			}
			return true;
		}

		/// <summary>
		/// ヒットテスト結果に応じたカーソルをセットする
		/// </summary>
		/// <param name="hitResilt">ヒットテスト結果</param>
		private void SetCursor(RegionalType hitResilt)
		{
			switch(hitResilt)
			{
			case RegionalType.Inner:
				System.Windows.Forms.Cursor.Current = Cursors.SizeAll;
				break;

			case RegionalType.TopLeft:
				System.Windows.Forms.Cursor.Current = Cursors.SizeNWSE;
				break;

			case RegionalType.TopRight:
				System.Windows.Forms.Cursor.Current = Cursors.SizeNESW;
				break;

			case RegionalType.BottomRight:
				System.Windows.Forms.Cursor.Current = Cursors.SizeNWSE;
				break;

			case RegionalType.BottomLeft:
				System.Windows.Forms.Cursor.Current = Cursors.SizeNESW;
				break;

			case RegionalType.TopEdge:
				System.Windows.Forms.Cursor.Current = Cursors.SizeNS;
				break;

			case RegionalType.RightEdge:
				System.Windows.Forms.Cursor.Current = Cursors.SizeWE;
				break;

			case RegionalType.BottomEdge:
				System.Windows.Forms.Cursor.Current = Cursors.SizeNS;
				break;

			case RegionalType.LeftEdge:
				System.Windows.Forms.Cursor.Current = Cursors.SizeWE;
				break;

			default:
				System.Windows.Forms.Cursor.Current = Cursors.Arrow;
				break;
			}
		}

		/// <summary>
		/// ピクチャーボックスのマウスイベント：移動
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _pictureBox_MouseMove(object sender, MouseEventArgs e)
		{
			RegionRectEditor rrs = _regionRectMap[SelectedRgnType];

			// マウスポインタ位置表示
			string str = string.Format("マウス(X={0}, Y={1})", e.X, e.Y);
			_toolStripStatusLabel_MousePoint.Text = str;

			// 編集中の領域情報表示
			if(rrs.IsMoving)
				str = string.Format("(X={0}, Y={1}, W={2}, H={3})", rrs.MovingRegionRect.Left, rrs.MovingRegionRect.Top, rrs.MovingRegionRect.Width, rrs.MovingRegionRect.Height);
			else
				str = string.Format("(X={0}, Y={1}, W={2}, H={3})", rrs.RegionRect.Left, rrs.RegionRect.Top, rrs.RegionRect.Width, rrs.RegionRect.Height);
			_toolStripStatusLabel_Msg.Text = str;

			// カーソル制御
			if(_hitTestResult == RegionalType.None)
			{
				SetCursor(rrs.HitTest(e.Location));
				return;
			}

			SetCursor(_hitTestResult);

			// マウスの移動量の計算
			Point mousePt = e.Location;

			Point delta = new Point(mousePt.X - _mouseButtonDownPoint.X, mousePt.Y - _mouseButtonDownPoint.Y);
			int dx = delta.X;
			int dy = delta.Y;

			rrs.StretchRegionRect(_hitTestResult, dx, dy);

			UpdatePictureBox();
		}

		/// <summary>
		/// ピクチャーボックスのマウスイベント：ボタンダウン
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _pictureBox_MouseDown(object sender, MouseEventArgs e)
		{
			if(e.Button == System.Windows.Forms.MouseButtons.Left)
			{
				RegionRectEditor selectedRRS = _regionRectMap[SelectedRgnType];

				_hitTestResult = selectedRRS.HitTest(e.Location);

				if(_hitTestResult != RegionalType.None)
				{
					SetCursor(_hitTestResult);

					selectedRRS.IsMoving = true;
					_mouseButtonDownPoint = e.Location;
				}
				else
				{
					// ポインタが選択されていない領域をクリックしたならば選択
					Func<SelectRgnType, bool> VlidateRRS = (type) =>
					{
						if(SelectedRgnType != type)
						{
							if(_regionRectMap[type].HitTest(e.Location) == RegionalType.Inner)
							{
								SelectedRgnType = type;
								UpdatePictureBox();
								
								return true;
							}
						}

						return false;
					};

					if(VlidateRRS(SelectRgnType.Artwork))
						return;
					if(VlidateRRS(SelectRgnType.Text))
						return;;
					if(VlidateRRS(SelectRgnType.BkImage))
						return;
				}
			}
		}

		/// <summary>
		/// ピクチャーボックスのマウスイベント：ボタンアップ
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void _pictureBox_MouseUp(object sender, MouseEventArgs e)
		{
			if(e.Button == System.Windows.Forms.MouseButtons.Left && _hitTestResult != RegionalType.None)
			{
				RegionRectEditor rrs = _regionRectMap[SelectedRgnType];

				if(rrs.IsMoving)
				{
					_hitTestResult = RegionalType.None;
					_mouseButtonDownPoint = new Point();

					rrs.IsMoving = false;

					switch(SelectedRgnType)
					{
					case SelectRgnType.Artwork:
						ArtworkDispRect = rrs.RegionRect;
						break;

					case SelectRgnType.Text:
						TextDispRect = rrs.RegionRect;
						break;

					case SelectRgnType.BkImage:
						BkImageDispRect = rrs.RegionRect;
						break;
					}
				}
			}
		}

		/// <summary>
		/// メニュー：領域アートワーク
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void artworkToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SelectedRgnType = SelectRgnType.Artwork;
			UpdatePictureBox();
		}

		/// <summary>
		/// メニュー：領域テキスト
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void textToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SelectedRgnType = SelectRgnType.Text;
			UpdatePictureBox();
		}

		/// <summary>
		/// メニュー：領域バックグランドイメージ
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void bkImageToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SelectedRgnType = SelectRgnType.BkImage;
			UpdatePictureBox();
		}

		/// <summary>
		/// メニュー：バックグランドイメージの領域をボードサイズに拡張する
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void fitBkImageSizeToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Rectangle rc = 	BkImageDispRect;
			rc.Location = new Point();
			rc.Size = BoardSize;
			BkImageDispRect = rc;

			_regionRectMap[SelectRgnType.BkImage].RegionRect = BkImageDispRect;
			UpdatePictureBox();
		}

		/// <summary>
		/// メニュー：OK
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void oKToolStripMenuItem_Click(object sender, EventArgs e)
		{
			DialogResult = System.Windows.Forms.DialogResult.OK;
			Close();
		}

		/// <summary>
		/// メニュー：キャンセル
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void cancelToolStripMenuItem_Click(object sender, EventArgs e)
		{
			DialogResult = System.Windows.Forms.DialogResult.Cancel;
			Close();
		}

		/// <summary>
		/// ディスクトップイメージの取得
		/// </summary>
		private void CaptureScreenImage()
		{
			if(_screenImage != null)
				_screenImage.Dispose();

			_screenImage = new Bitmap(Screen.PrimaryScreen.Bounds.Width, Screen.PrimaryScreen.Bounds.Height);

			using(Graphics g = Graphics.FromImage(_screenImage))
			{
				g.CopyFromScreen(new Point(), new Point(), _screenImage.Size);
			}
		}

		/// <summary>
		/// 表示位置が変更された場合は背景を更新するためにピクチャーボックスをアップデータする
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void FormRegionSelect_LocationChanged(object sender, EventArgs e)
		{
			UpdatePictureBox();
		}

		
	}
}
