﻿using System;
using System.Collections.Generic;
using System.Text;

using System.Runtime.InteropServices;
using System.IO;

namespace NTag.MP3.ID3v1
{
	/// <summary>
	/// ID3v1 読み込みクラス
	/// </summary>
	sealed public class ID3v1Reader : ITagReader
	{
		/// <summary>
		/// TAG サイズ（128 固定）
		/// </summary>
		private static readonly int TAGSIZE = 128;

		/// <summary>
		/// ジャンルテキスト1 0x00 - 0x93
		/// (Id3Lib より定義を流用しました)
		/// </summary>
		private static readonly string[] _genres1 =
		{
			"Blues","Classic Rock","Country","Dance","Disco","Funk","Grunge","Hip-Hop","Jazz","Metal",
			"New Age","Oldies","Other","Pop","R&B","Rap","Reggae","Rock","Techno","Industrial",
			"Alternative","Ska","Death Metal","Pranks","Soundtrack","Euro-Techno","Ambient","Trip-Hop",
			"Vocal","Jazz+Funk","Fusion","Trance","Classical","Instrumental","Acid","House",
			"Game","Sound Clip","Gospel","Noise","Alternative Rock","Bass","Soul","Punk","Space",
			"Meditative","Instrumental Pop","Instrumental Rock","Ethnic","Gothic",
			"Darkwave","Techno-Industrial","Electronic","Pop-Folk","Eurodance","Dream",
			"Southern Rock","Comedy","Cult","Gangsta","Top 40","Christian Rap","Pop/Funk","Jungle",
			"Native American","Cabaret","New Wave","Psychadelic","Rave","Showtunes","Trailer","Lo-Fi",
			"Tribal","Acid Punk","Acid Jazz","Polka","Retro","Musical","Rock & Roll","Hard Rock","Folk",
			"Folk/Rock","National Folk","Swing","Fast-Fusion","Bebob","Latin","Revival","Celtic","Bluegrass",
			"Avantgarde","Gothic Rock","Progressive Rock","Psychedelic Rock","Symphonic Rock","Slow Rock",
			"Big Band","Chorus","Easy Listening","Acoustic","Humour","Speech","Chanson","Opera","Chamber Music",
			"Sonata","Symphony","Booty Bass","Primus","Porn Groove","Satire","Slow Jam","Club",
			"Tango","Samba","Folklore","Ballad","Power Ballad","Rhytmic Soul","Freestyle","Duet",
			"Punk Rock","Drum Solo","Acapella","Euro-House","Dance Hall","Goa","Drum & Bass","Club-House",
			"Hardcore","Terror","Indie","BritPop","Negerpunk","Polsk Punk","Beat","Christian Gangsta Rap",
			"Heavy Metal","Black Metal","Crossover","Contemporary Christian",
			"Christian Rock","Merengue","Salsa","Trash Metal","Anime","JPop","SynthPop"
		};

		/// <summary>
		/// ジャンルテキスト2 0xf0 - 0xff
		/// </summary>
		private static readonly string[] _genres2 =
		{
			"Sacred", "Northern Europe", "Irish & Scottish",  "Scotland", "Ethnic Europe", "Enka", "Children's Song", "",
			"Heavy Rock(J)", "Doom Rock(J)", "J-POP(J)", "Seiyu(J)", "Tecno Ambient(J)", "Moemoe(J)", "Tokusatsu(J)", "Anime(J)"
		};

		/// <summary>
		/// タグデータ
		/// </summary>
		public Tag Tag { private set; get; }

		/// <summary>
		/// タグを読み込むファイルのパスの取得
		/// </summary>
		public string FilePath { private set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="filePath">タグを読み込むファイルのパス</param>
		public ID3v1Reader(string filePath)
		{
			FilePath = filePath;
		}

		/// <summary>
		/// 有効な ID3V1 ファイルであるかを確認します
		/// </summary>
		/// <returns></returns>
		public bool IsValidFile()
		{
			ID3V10 id3v10 = null;
			ID3V11 id3v11 = null;

			if(!ReadData(FilePath, out id3v10, out id3v11))
				return false;

			if(!Function.ByteArrayCompare(id3v10.ID, new byte[] { 0x54, 0x41, 0x47 }))
				return false;

			return true;
		}

		/// <summary>
		/// タグ読み込み
		/// </summary>
		public void Read()
		{
			Encoding enc = Encoding.GetEncoding("shift_jis");
			
			Tag = null;

			try
			{
				// タグ読込み
				ID3V10 id3v10 = null;
				ID3V11 id3v11 = null;

				if(!ReadData(FilePath, out id3v10, out id3v11))
					throw new NTagException(FilePath + " は ID3V1 ファイルではありません。");

				// タグへ変換
				Tag = new Tag();
				Tag.FilePath = FilePath;
				Tag.TagType = NTag.TagType.ID3V1;

				Tag.Title = GetString(enc, id3v10.Title);
				Tag.Artist = GetString(enc, id3v10.Artist);
				Tag.Album = GetString(enc, id3v10.Album);
				Tag.Year = GetString(enc, id3v10.Year);

				if(id3v11.TrackNumber[0] == 0)
				{
					Tag.Comment = GetString(enc, id3v11.Comment);
					Tag.TrackNumber1 = (int)id3v11.TrackNumber[1];
					Tag.TrackNumberString = Tag.TrackNumber1.ToString();
				}
				else
					Tag.Comment = GetString(enc, id3v10.Comment);

				if(id3v10.Genre <= 0x93)
					Tag.Genre = _genres1[id3v10.Genre];
				else if(id3v10.Genre >= 0xf0)
					Tag.Genre = _genres2[id3v10.Genre - 0xf0];
			}
			catch
			{
				Tag = null;
				throw;
			}
		}

		/// <summary>
		/// 0x20 のパティングを除去して文字列へ変換する
		/// </summary>
		/// <param name="enc">文字エンコード</param>
		/// <param name="data">変換するバイト配列</param>
		/// <returns></returns>
		private string GetString(Encoding enc, byte[] data)
		{
			int strCount = 0;

			for(int i = data.Length - 1 ; i >= 0 ; i--)
			{
				if(data[i] != 0x20)
				{
					strCount = i + 1;
					break;
				}
			}

			if(strCount <= 0)
				return "";

			return enc.GetString(data, 0, strCount);
		}

		/// <summary>
		/// データを読み込む
		/// </summary>
		/// <param name="filePath">ファイルパス</param>
		/// <param name="id3v10">ID3V1.0 用の構造体</param>
		/// <param name="id3v11">ID3V1.1 用の構造体</param>
		/// <returns></returns>
		private static bool ReadData(string filePath, out ID3V10 id3v10, out ID3V11 id3v11)
		{
			byte[] data = new byte[TAGSIZE];

			id3v10 = null;
			id3v11 = null;

			// データリード
			using(FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read))
			{
				// ファイルのサイズをチェック
				if(fs.Length < TAGSIZE)
					return false;

				fs.Seek(-TAGSIZE, SeekOrigin.End);
				fs.Read(data, 0, TAGSIZE);
			}

			// 構造体にマーシャリング
			GCHandle gch = GCHandle.Alloc(data, GCHandleType.Pinned);
			try
			{
				id3v10 = (ID3V10)Marshal.PtrToStructure(gch.AddrOfPinnedObject(), typeof(ID3V10));
				id3v11 = (ID3V11)Marshal.PtrToStructure(gch.AddrOfPinnedObject(), typeof(ID3V11));
			}
			finally
			{
				gch.Free();
			}

			return true;
		}

		/// <summary>
		/// ID3V1.1 用の構造体
		/// </summary>
		[StructLayout(LayoutKind.Sequential, Pack = 4, CharSet = CharSet.Ansi)]
		private class ID3V11
		{
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
			public byte[] ID;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Title;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Artist;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Album;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
			public byte[] Year;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 28)]
			public byte[] Comment;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 2)]
			public byte[] TrackNumber;
			
			public byte Genre;
		};

		/// <summary>
		/// ID3V1.0 用の構造体
		/// </summary>
		[StructLayout(LayoutKind.Sequential, Pack = 4, CharSet = CharSet.Ansi)]
		private class ID3V10
		{
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
			public byte[] ID;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Title;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Artist;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Album;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
			public byte[] Year;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 30)]
			public byte[] Comment;
			
			public byte Genre;
		};
	}
}
