﻿using System;
using System.Collections.Generic;
using System.Text;

using System.IO;

namespace NTag.MP3.ID3v2
{
	/// <summary>
	/// ID3v2 共通の処理、データの定義クラス
	/// </summary>
	sealed class ID3v2
	{
		/// <summary>
		/// ID3v2 ヘッダフラグ
		/// </summary>
		[Flags]
		public enum ID3v2HeaderFlags : byte
		{
			/// <summary>
			/// なし（初期値）
			/// </summary>
			None = 0x00,

			/// <summary>
			/// 非同期化
			/// </summary>
			Unsync = 0x80,
			/// <summary>
			/// 拡張ヘッダあり
			/// </summary>
			ExHeader = 0x40,
			/// <summary>
			/// 実験中
			/// </summary>
			Testing = 0x20,
		}

		/// <summary>
		/// フレームタイプ
		/// </summary>
		public enum FrameTpey
		{
			/// <summary>
			/// なし（初期値）
			/// </summary>
			None,
			/// <summary>
			/// テキスト
			/// </summary>
			Text,
			/// <summary>
			/// イメージ
			/// </summary>
			Picture,
			/// <summary>
			/// バイナリ
			/// </summary>
			Binary,
		}

		/// <summary>
		/// ID3v3 フレームヘッダフラグ
		/// </summary>
		[Flags]
		public enum FreamHeaderFalgs : ushort
		{
			/// <summary>
			/// なし（初期値）
			/// </summary>
			None = 0x0000,

			/// <summary>
			///  タグの変更で取り除かれる
			/// </summary>
			RemoveTagChange = 0x8000,
			/// <summary>
			/// ファイルの変更で取り除かれる 
			/// </summary>
			RemoveFileChange = 0x4000,
			/// <summary>
			/// 読み込み専用
			/// </summary>
			ReadOnly = 0x2000,

			/// <summary>
			/// 圧縮されている
			/// </summary>
			Compression = 0x0080,
			/// <summary>
			/// 暗号化されている
			/// </summary>
			Encryption = 0x0040,
			/// <summary>
			/// グループ識別子
			/// </summary>
			Grouping = 0x0020,
		}

		/// <summary>
		/// 文字データの文字コード
		/// </summary>
		public enum TextCode : byte
		{
			/// <summary>
			/// 文字コードはない
			/// </summary>
			None = 0xff,

			/// <summary>
			/// ISO-8859-1 [ISO-8859-1]。終端文字は $00
			/// </summary>
			ISO_8859_1 = 0x00,
			/// <summary>
			/// [UTF-16] でエンコードされた Unicode [UNICODE]。BOM有り
			/// 一つのフレームの内部では、すべて同じバイトオーダーであるべきである
			/// 終端文字は $00 00
			/// </summary>
			UTF16 = 0x01,
			/// <summary>
			/// ビッグエンディアン）でエンコードされた Unicode
			/// BOMなし
			/// 終端文字は $00 00
			/// </summary>
			UTF16BE = 0x02,
			/// <summary>
			///  [UTF-8] でエンコードされた Unicode
			///  終端文字は $00
			/// </summary>
			UTF8 = 0x03,
		}

		/// <summary>
		/// イメージタイプ（フォーマット）(MIME Type)
		/// </summary>
		public enum ImageType
		{
			/// <summary>
			/// なし (不明なフォーマット)
			/// </summary>
			None,

			/// <summary>
			/// JPEG (image/jpg or image/jpeg)
			/// </summary>
			Jpeg,
			/// <summary>
			/// PNG (image/png)
			/// </summary>
			Png,
		}

		/// <summary>
		/// 画像タイプ
		/// </summary>
		public enum PictureType : byte
		{
			/// <summary>
			/// Other
			/// </summary>
			Other = 0x00,
			/// <summary>
			/// 32x32 pixels 'file icon' (PNG only)
			/// </summary>
			FileIcon32x32 = 0x01,
			/// <summary>
			/// Other file icon
			/// </summary>
			FileIconOther = 0x02,
			/// <summary>
			/// Cover (front)
			/// </summary>
			CoverFront = 0x03,
			/// <summary>
			/// Cover (back)
			/// </summary>
			CoverBack = 0x04,
			/// <summary>
			/// Leaflet page
			/// </summary>
			LeafletPage = 0x05,
			/// <summary>
			/// Media (e.g. lable side of CD)
			/// </summary>
			Media = 0x06,
			/// <summary>
			/// Lead artist/lead performer/soloist
			/// </summary>
			LeadArtist = 0x07,
			/// <summary>
			/// Artist/performer
			/// </summary>
			Artist = 0x08,
			/// <summary>
			/// Conductor
			/// </summary>
			Conductor = 0x09,
			/// <summary>
			/// Band/Orchestra
			/// </summary>
			Band = 0x0a,
			/// <summary>
			/// Composer
			/// </summary>
			Composer = 0x0b,
			/// <summary>
			/// Lyricist/text writer
			/// </summary>
			Lyricist = 0x0c,
			/// <summary>
			/// Recording Location
			/// </summary>
			RecordingLocation = 0x0d,
			/// <summary>
			/// During recording
			/// </summary>
			DuringRecording = 0x0e,
			/// <summary>
			/// During performance
			/// </summary>
			DuringPerformance = 0x0f,
			/// <summary>
			/// Movie/video screen capture
			/// </summary>
			ScreenCapture = 0x10,
			/// <summary>
			/// A bright coloured fish
			/// </summary>
			BrightColouredFish = 0x11,
			/// <summary>
			/// Illustration
			/// </summary>
			Illustration = 0x12,
			/// <summary>
			/// Band/artist logotype
			/// </summary>
			BandLogotype = 0x13,
			/// <summary>
			/// Publisher/Studio logotype
			/// </summary>
			PublisherLogotype = 0x14,
		}

		/// <summary>
		/// 文字コードに対応する文字エンコーダーを返す
		/// </summary>
		/// <returns></returns>
		public static Encoding GetEncoding(TextCode code)
		{
			// 文字データの判別
			switch(code)
			{
			case TextCode.ISO_8859_1:
				return Encoding.GetEncoding("shift_jis");

			case TextCode.UTF16:
				return Encoding.Unicode;

			case TextCode.UTF16BE:
				return Encoding.BigEndianUnicode;

			case TextCode.UTF8:
				return Encoding.UTF8;
			}

			return Encoding.Default;
		}

		/// <summary>
		/// 文字コードを取得する
		/// </summary>
		/// <param name="data"></param>
		/// <returns></returns>
		public static TextCode GetTextCode(byte data)
		{
			// 文字コードの判別
			switch(data)
			{
			case 0:
				return ID3v2.TextCode.ISO_8859_1;

			case 1:
				return ID3v2.TextCode.UTF16;

			case 2:
				return ID3v2.TextCode.UTF16BE;

			case 3:
				return ID3v2.TextCode.UTF8;
			}

			return ID3v2.TextCode.None;
		}
	}
}
