﻿using System;
using System.Collections.Generic;
using System.Text;

using System.IO;

namespace NTag.MP3.ID3v2
{
	/// <summary>
	/// ID3v2 ヘッダクラス
	/// </summary>
	sealed class ID3v2Header
	{
		/// <summary>
		/// ヘッダサイズ（10 固定）
		/// </summary>
		private static readonly int HEADERSIZE = 10;

		/// <summary>
		/// ヘッダのバイトデータ
		/// </summary>
		public byte[] HeaderData { set; get; }

		/// <summary>
		/// IDのヘッダータグ 正常ならば "ID3"
		/// </summary>
		public byte[] ID { set; get; }

		/// <summary>
		/// メジャーバージョン
		/// </summary>
		public int MajarVer { set; get; }
		/// <summary>
		/// マイナーバージョン
		/// </summary>
		public int MinorVer { set; get; }
		/// <summary>
		/// フラグ
		/// </summary>
		public ID3v2.ID3v2HeaderFlags Flags { set; get; }

		/// <summary>
		/// タグのデータサイズ[byte]
		/// （タグの全サイズはヘッダ 10 バイトを加えたサイズになる。TagSize を使用するといい）
		/// </summary>
		public int Size { set; get; }
		/// <summary>
		/// ヘッダを含めたタグのサイズ[byte]
		/// </summary>
		public int TagSize { get { return Size + HEADERSIZE; } }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		public ID3v2Header()
		{
			HeaderData = new byte[HEADERSIZE];
			ID = new byte[3];
			MajarVer = 0;
			MinorVer = 0;

			Flags = ID3v2.ID3v2HeaderFlags.None;

			Size = 0;
		}

		/// <summary>
		/// ヘッダが正常な ID3v2 ヘッダかを調べます
		/// </summary>
		/// <returns></returns>
		public bool IsValid()
		{
			// "ID3" を確認
			if(!Function.ByteArrayCompare(ID, new byte[] { 0x49, 0x44, 0x33 }))
				return false;

			// バージョンを確認
			if(MajarVer != 3 && MajarVer != 4)
				return false;

			// 非同期タグはサポートしない
			if((Flags & ID3v2.ID3v2HeaderFlags.Unsync) == ID3v2.ID3v2HeaderFlags.Unsync)
				return false;
			// テスト中はサポートしない
			if((Flags & ID3v2.ID3v2HeaderFlags.Testing) == ID3v2.ID3v2HeaderFlags.Testing)
				return false;
			// 拡張ヘッダはサポートしない
			if((Flags & ID3v2.ID3v2HeaderFlags.ExHeader) == ID3v2.ID3v2HeaderFlags.ExHeader)
				return false;

			return true;
		}

		/// <summary>
		/// ヘッダを読み込む
		/// </summary>
		/// <param name="fs">有効なファイルストリーム</param>
		/// <returns>読み込んだバイト数</returns>
		public int Read(FileStream fs)
		{
			int readBytes = 0;

			fs.Seek(0, 0);
			readBytes = fs.Read(HeaderData, 0, HEADERSIZE);
			if(readBytes != HEADERSIZE)
				return -1;

			// ID 取得
			Buffer.BlockCopy(HeaderData, 0, ID, 0, 3);

			// バージョン取得
			MajarVer = (int)HeaderData[3];
			MinorVer = (int)HeaderData[4];

			// サイズ取得
			Flags = (ID3v2.ID3v2HeaderFlags)(HeaderData[5]);

			{
				int b1 = (int)(HeaderData[6] & 0x7f) << 21;
				int b2 = (int)(HeaderData[7] & 0x7f) << 14;
				int b3 = (int)(HeaderData[8] & 0x7f) << 7;
				int b4 = (int)(HeaderData[9] & 0x7f);

				Size = b1 | b2 | b3 | b4;
			}

			if(Size <= 0)
				return -1;
			if(Size >= (fs.Length - fs.Position))
				return -1;

			return HEADERSIZE;
		}
	}
}
