﻿using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Drawing;

namespace NTag.MP3.ID3v2
{
	/// <summary>
	/// ID3v2 読み込みクラス
	/// </summary>
	sealed public class ID3v2Reader : ITagReader
	{
		/// <summary>
		/// タグデータ
		/// </summary>
		public Tag Tag { private set; get; }

		/// <summary>
		/// タグを読み込むファイルのパスの取得
		/// </summary>
		public string FilePath { private set; get; }

		/// <summary>
		/// テキストフレームデータ
		/// </summary>
		private List<Frame.Frame> _textFrames = new List<Frame.Frame>();
		/// <summary>
		/// 画像フレームデータ
		/// </summary>
		private List<Frame.Frame> _pictureFrames = new List<Frame.Frame>();

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="filePath">タグを読み込むファイルのパス</param>
		public ID3v2Reader(string filePath)
		{
			FilePath = filePath;
		}

		/// <summary>
		/// 有効な ID3V2 ファイルであるかを確認します
		/// </summary>
		/// <returns></returns>
		public bool IsValidFile()
		{
			ID3v2Header id3v2Header = new ID3v2Header();

			using(FileStream fs = new FileStream(FilePath, FileMode.Open, FileAccess.Read))
			{
				id3v2Header.Read(fs);
				return id3v2Header.IsValid();
			}
		}

		/// <summary>
		/// タグ読み込み
		/// </summary>
		public void Read()
		{
			Tag = null;

			_textFrames.Clear();
			_pictureFrames.Clear();

			try
			{
				///////////////////////////////////////////////////////////////////
				// 全フレームの読込み
				ID3v2Header id3v2Header = new ID3v2Header();

				int readBytes = 0;
				int offset = 0;

				using(FileStream fs = new FileStream(FilePath, FileMode.Open, FileAccess.Read))
				{
					// ヘッダ読込み
					readBytes = id3v2Header.Read(fs);
					if(readBytes < 0)
						throw new NTagException(FilePath + " は ID3V2 ファイルではありません。");

					offset = readBytes;

					// ID3V2 であることを確認する
					if(!id3v2Header.IsValid())
						throw new NTagException(FilePath + " は ID3V2 ファイルではありません。");

					Tag = new Tag();
					Tag.FilePath = FilePath;

					if(id3v2Header.MajarVer == 3)
						Tag.TagType = NTag.TagType.ID3V2_3;
					else if(id3v2Header.MajarVer == 4)
						Tag.TagType = NTag.TagType.ID3V2_4;
					else
						throw new NTagException(FilePath + " は ID3V2 ファイルではありません。");

					do
					{
						Frame.Frame frame;
						readBytes = Frame.Frame.Read(fs, offset, out frame);

						if(readBytes < 0)
							break;

						if(frame != null)
						{
							switch(frame.FrameTpey)
							{
							case ID3v2.FrameTpey.Text:
								_textFrames.Add(frame);
								break;

							case ID3v2.FrameTpey.Picture:
								_pictureFrames.Add(frame);
								break;
							}
						}

						offset += readBytes;

					} while(offset < id3v2Header.TagSize);
				}

				///////////////////////////////////////////////////////////////////
				// フレーム解析
				foreach(Frame.Frame frame in _textFrames)
				{
					Frame.TextFrame testFrame = (Frame.TextFrame)frame;

					if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Title])
						Tag.Title = testFrame.GetText();
					else if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Artist])
						Tag.Artist = testFrame.GetText();
					else if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Album])
						Tag.Album = testFrame.GetText();
					else if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Year])
						Tag.Year = testFrame.GetText();
					else if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Genre])
						Tag.Genre = testFrame.GetText();
					else if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Comment])
						Tag.Comment = testFrame.GetCommentText();
					else if(testFrame.FreamHeader.ID == Frame.Frame.TEXTDATA_FRAMEID_TABLE[NTag.TagFieldFlags.TrackNumber1])
					{
						Tag.TrackNumberString = testFrame.GetText();

						int num;
						if(int.TryParse(Tag.TrackNumberString, out num))
							Tag.TrackNumber1 = num;
					}
				}

				foreach(Frame.Frame frame in _pictureFrames)
				{
					Frame.PictureFrame pictureFrame = (Frame.PictureFrame)frame;

					if(pictureFrame.FreamHeader.ID == Frame.Frame.PICDATA_FRAMEID_TABLE[NTag.TagFieldFlags.Artwork])
					{
						Image image = pictureFrame.GetPicture();

						if(image != null)
							Tag.Artworks.Add(image);
					}
				}

				if(Tag.Artworks.Count > 0)
				{
					Tag.Artwork = new Bitmap(Tag.Artworks[0]);
				}
			}
			catch
			{
				Tag = null;
				throw;
			}
		}
	}
}
