﻿using System;
using System.Collections.Generic;

using System.Text;
using System.Drawing;
using System.IO;

// 参考にしたHP : http://atomicparsley.sourceforge.net/mpeg-4files.html

namespace NTag.MP4
{
	/// <summary>
	/// AAC のタグを読み込むクラス
	/// </summary>
	public sealed class AACTagReader : ITagReader
	{
		/// <summary>
		/// BOX データの読み込みのステート
		/// </summary>
		private enum BoxReadState
		{
			Search_moov,
			Search_udta,
			Search_meta,
			Finish,
		}

		/// <summary>
		/// タグデータ
		/// </summary>
		public Tag Tag { private set; get; }

		/// <summary>
		/// タグを読み込むファイルのパスの取得
		/// </summary>
		public string FilePath { private set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="filePath">タグを読み込むファイルのパス</param>
		public AACTagReader(string filePath)
		{
			FilePath = filePath;
		}

		/// <summary>
		/// 有効な AAC ファイルであるかを確認します
		/// </summary>
		/// <returns></returns>
		public bool IsValidFile()
		{
			using(FileStream fs = new FileStream(FilePath, FileMode.Open, FileAccess.Read))
			{
				return IsValidAACFile(fs);
			}
		}

		/// <summary>
		/// 有効な AAC ファイルであるかを確認します
		/// </summary>
		/// <param name="fs">有効なファイルストリーム</param>
		/// <returns></returns>
		private bool IsValidAACFile(FileStream fs)
		{
			byte[] tmp = new byte[12];

			try
			{
				fs.Seek(0, 0);
				fs.Read(tmp, 0, 12);
				
				// ftyp ?
				if(!Function.ByteArrayCompare(tmp, 4, new byte[] { 0x66, 0x74, 0x79, 0x70 }, 0, 4))
					return false;

				// "M4A " ?
				if(Function.ByteArrayCompare(tmp, 8, new byte[] { 0x4d, 0x34, 0x41, 0x20 }, 0, 4))
					return true;

				// "isom" ?
				if(Function.ByteArrayCompare(tmp, 8, new byte[] { 0x69, 0x73, 0x6f, 0x6d }, 0, 4))
					return true;

				return false;
			}
			finally
			{
				fs.Seek(0, 0);
			}
		}

		/// <summary>
		/// タグ読み込み
		/// </summary>
		public void Read()
		{
			Tag = null;

			BoxReadState readState = BoxReadState.Search_moov;

			try
			{
				using(FileStream fs = new FileStream(FilePath, FileMode.Open, FileAccess.Read))
				{
					int readBytes = 0;
					int offset = 0;
					int limit = (int)fs.Length;

					// AAC であることを確認する
					if(!IsValidAACFile(fs))
						return;

					// MPEG4 の仕様に従って "meta" を読み込む

					BoxData boxData = new BoxData();

					do
					{
						// BOX 読込み
						readBytes = boxData.Read(fs, offset);
						if(readBytes < 0)
							return;

						switch(readState)
						{
						// ① moov BOX が見つかった場合、次に入れ子の udat を探す
						case BoxReadState.Search_moov:
							
							if(boxData.TypeStr == "moov")
							{
								limit = offset + boxData.Size;
								offset += 8;

								readState = BoxReadState.Search_udta;
							}
							else
								offset += boxData.Size;

							break;

						// ② udta BOX が見つかった場合、次に入れ子の meta を探す
						case BoxReadState.Search_udta:

							if(boxData.TypeStr == "udta")
							{
								limit = offset + boxData.Size;
								offset += 8;

								readState = BoxReadState.Search_meta;
							}
							else
								offset += boxData.Size;

							break;

						// ③ meta BOX が見つかった場合、meta に TAG 情報が入っているのでデータを解析
						case BoxReadState.Search_meta:

							if(boxData.TypeStr == "meta")
							{
								AnalyzeMetaData(boxData.Data);

								readState = BoxReadState.Finish;
							}
							else
								offset += boxData.Size;

							break;
						}

						if(readState == BoxReadState.Finish)
							break;

					} while(offset < limit);
				}
			}
			catch
			{
				Tag = null;
				throw;
			}
		}

		
		/// <summary>
		/// メタデータの解析
		/// </summary>
		/// <param name="data"></param>
		/// <returns></returns>
		private void AnalyzeMetaData(byte[] data)
		{
			Atom.AtomBasic ilstAtom = null;

			// meta 内のアトムを読み込み、ilst アトムを探す
			{
				List<Atom.AtomBasic> atoms = ReadAtoms(data);

				foreach(Atom.AtomBasic atom in atoms)
				{
					if(Function.ByteArrayCompare(atom.Header.Name, Atom.Atom.ILST_ATOMNAME))
					{
						if(atom.IsHaveChild())
						{
							ilstAtom = atom;
							break;
						}
					}
				}
			}

			// ilst アトムが見つからない場合は終了
			if(ilstAtom == null)
				return;

			// ilst 内のアトムを読み込みタグデータを取得する
			try
			{
				Tag = new Tag();
				Tag.FilePath = FilePath;
				Tag.TagType = NTag.TagType.AAC;

				List<Atom.AtomBasic> atoms = ReadAtoms(ilstAtom.Data);

				foreach(Atom.AtomBasic atom in atoms)
				{
					Atom.AtomData atomData = null;

					if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Title)) != null)
						Tag.Title = atomData.GetText();
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Artist)) != null)
						Tag.Artist = atomData.GetText();
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Album)) != null)
						Tag.Album = atomData.GetText();
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Genre)) != null)
						Tag.Genre = atomData.GetText();
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Year)) != null)
						Tag.Year = atomData.GetText();
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Comment)) != null)
						Tag.Comment = atomData.GetText();
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.TrackNumber1)) != null)
					{
						Tag.TrackNumber1 = atomData.Data[3];
						Tag.TrackNumber2 = atomData.Data[5];

						Tag.TrackNumberString = string.Format("{0:D2}/{1:D2}", Tag.TrackNumber1, Tag.TrackNumber2);
					}
					else if((atomData = ReadAtomData(atom, NTag.TagFieldFlags.Artwork)) != null)
					{
						Tag.Artwork = atomData.GetPicture();
						Tag.Artworks.Add(new Bitmap(Tag.Artwork));
					}
				}
			}
			catch
			{
				Tag = null;
				throw;
			}
		}

		/// <summary>
		/// アトム内の data アトムを読み込む
		/// （存在しない場合は null）
		/// </summary>
		/// <param name="atom"></param>
		/// <param name="flag"></param>
		/// <returns></returns>
		private Atom.AtomData ReadAtomData(Atom.AtomBasic atom, NTag.TagFieldFlags flag)
		{
			if(Function.ByteArrayCompare(atom.Header.Name, Atom.Atom.ATOM_NAME_TABLE[flag]))
			{
				if(atom.IsHaveChild())
				{
					List<Atom.AtomBasic> dataAtoms = ReadAtoms(atom.Data);

					foreach(Atom.AtomBasic dataAtom in dataAtoms)
					{
						if(Function.ByteArrayCompare(dataAtom.Header.Name, Atom.Atom.DATA_ATOMNAME))
						{
							if(!dataAtom.IsHaveChild())
							{
								Atom.AtomData data = new Atom.AtomData(dataAtom);
								return data;
							}
						}
					}
				}
			}

			return null;
		}

		/// <summary>
		/// data から Atom の一覧を取得する
		/// </summary>
		/// <param name="data"></param>
		/// <returns></returns>
		private List<Atom.AtomBasic> ReadAtoms(byte[] data)
		{
			int offset = 0;
			int limit = data.Length;

			System.Text.Encoding enc = System.Text.Encoding.UTF8;

			// データの先頭が 0x00000000 ならば読み飛ばす (4バイトの NULL スペース)
			if(data[offset] == 0 && data[offset + 1] == 0 && data[offset + 2] == 0 && data[offset + 3] == 0)
				offset += 4;

			// アトムを読み込んでいく
			List<Atom.AtomBasic> atoms = new List<Atom.AtomBasic>();

			do
			{
				Atom.AtomBasic atom = new Atom.AtomBasic();

				int readBytes = atom.Read(data, offset);
				atoms.Add(atom);

				offset += readBytes;
			} while(offset < limit);

			return atoms;
		}
		
	}
}