﻿using System;
using System.Collections.Generic;
using System.Text;

using System.IO;

namespace NTag
{
	/// <summary>
	/// MP3 (ID3v1, ID3v2), MP4 (AAC) タグの読込みクラス
	/// </summary>
	public sealed class TagReader : ITagReader
	{
		/// <summary>
		/// ファイル拡張子でタグリーダーを選択する際のテーブル
		/// （キーの拡張子は小文字 ".mp3"、値のタグタイプは優先順位の高い順）
		/// </summary>
		public Dictionary<string, List<NTag.TagType>> ExtensionTable { private set; get; }

		/// <summary>
		/// タグデータ
		/// </summary>
		public Tag Tag { private set; get; }

		/// <summary>
		/// タグを読み込むファイルのパスの取得
		/// </summary>
		public string FilePath { private set; get; }

		/// <summary>
		/// タグリーダー選択方法
		/// </summary>
		public NTag.ReadSelectType ReadSelectType { set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="filePath">タグを読み込むファイルのパス</param>
		public TagReader(string filePath)
		{
			FilePath = filePath;
			ReadSelectType = NTag.ReadSelectType.Extension;

			ExtensionTable = new Dictionary<string, List<NTag.TagType>>();
			ExtensionTable.Add(".mp3", new List<NTag.TagType> { NTag.TagType.ID3V2_x, NTag.TagType.ID3V1 });
			ExtensionTable.Add(".m4a", new List<NTag.TagType> { NTag.TagType.AAC });
			ExtensionTable.Add(".aac", new List<NTag.TagType> { NTag.TagType.AAC });

		}

		/// <summary>
		/// 有効なタグファイルであるかを確認します
		/// </summary>
		/// <returns></returns>
		public bool IsValidFile()
		{
			ITagReader reader = GetTagReader();

			if(reader == null)
				return false;

			return true;
		}

		/// <summary>
		/// タグリーダークラスを取得します。
		/// タグがない場合は null を返します。
		/// </summary>
		/// <returns></returns>
		private ITagReader GetTagReader()
		{
			ITagReader reader = null;

			if(ReadSelectType == NTag.ReadSelectType.FileStruct)
			{
				// MP3 ?
				reader = new MP3.MP3TagReader(FilePath);
				if(!reader.IsValidFile())
				{
					// AAC ?
					reader = new MP4.AACTagReader(FilePath);
					if(!reader.IsValidFile())
					{
						// タグなし
						return null;
					}
				}
			}
			else
			{
				string ext = Path.GetExtension(FilePath).ToLower();

				if(ExtensionTable.ContainsKey(ext))
				{
					List<NTag.TagType> tagTypes = ExtensionTable[ext];

					foreach(NTag.TagType tagType in tagTypes)
					{
						if((tagType & NTag.TagType.AAC) == NTag.TagType.AAC)
						{
							reader = new MP4.AACTagReader(FilePath);
							if(reader.IsValidFile())
								break;
						}
						else if((tagType & NTag.TagType.ID3V1) == NTag.TagType.ID3V1)
						{
							reader = new MP3.ID3v1.ID3v1Reader(FilePath);
							if(reader.IsValidFile())
								break;
						}
						else if((tagType & NTag.TagType.ID3V2_x) == NTag.TagType.ID3V2_x)
						{
							reader = new MP3.ID3v2.ID3v2Reader(FilePath);
							if(reader.IsValidFile())
								break;
						}
					}
				}
			}

			return reader;
		}

		/// <summary>
		/// タグ読み込み
		/// </summary>
		/// <param name="filePath">読み込むファイル名</param>
		public void Read()
		{
			Tag = null;

			try
			{
				ITagReader reader = GetTagReader();

				if(reader == null)
					return;

				reader.Read();
				Tag = reader.Tag;
			}
			catch
			{
				Tag = null;
				throw;
			}
		}
	}
}
