﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Drawing;

namespace NauPlaying
{
	/// <summary>
	/// 頂点座標で管理する Rectangle 構造体
	/// </summary>
	public struct RectangleV
	{
		/// <summary>
		/// 左上 X 座標
		/// </summary>
		private int _x1;
		/// <summary>
		/// 左上 Y 座標
		/// </summary>
		private int _y1;
		/// <summary>
		/// 右下 X 座標
		/// </summary>
		private int _x2;
		/// <summary>
		/// 右下 Y2 座標
		/// </summary>
		private int _y2;

		/// <summary>
		/// 左上 X 座標の設定、取得
		/// </summary>
		public int X1
		{
			get { return _x1; }
			set { _x1 = value; }
		}
		/// <summary>
		/// 左上 Y 座標の設定、取得
		/// </summary>
		public int Y1
		{
			get { return _y1; }
			set { _y1 = value; }
		}
		/// <summary>
		/// 右下 X 座標の設定、取得
		/// </summary>
		public int X2
		{
			get { return _x2; }
			set { _x2 = value; }
		}
		/// <summary>
		/// 右下 Y2 座標の設定、取得
		/// </summary>
		public int Y2
		{
			get { return _y2; }
			set { _y2 = value; }
		}

		/// <summary>
		/// 幅の取得
		/// </summary>
		public int Width
		{
			get { return _x2 - _x1 + 1; }
		}
		/// <summary>
		/// 高さの取得
		/// </summary>
		public int Height
		{
			get { return _y2 - _y1 + 1; }
		}

		/// <summary>
		/// 左上 X 座標の設定、取得
		/// </summary>
		public int Left
		{
			get { return _x1; }
			set { _x1 = value; }
		}
		/// <summary>
		/// 左上 Y 座標の設定、取得
		/// </summary>
		public int Top
		{
			get { return _y1; }
			set { _y1 = value; }
		}
		/// <summary>
		/// 右下 X 座標の設定、取得
		/// </summary>
		public int Right
		{
			get { return _x2; }
			set { _x2 = value; }
		}
		/// <summary>
		/// 右下 Y2 座標の設定、取得
		/// </summary>
		public int Bottom
		{
			get { return _y2; }
			set { _y2 = value; }
		}

		/// <summary>
		/// 矩形サイズの取得
		/// </summary>
		public Size Size
		{
			get { return new Size(Width, Height); }
		}

		/// <summary>
		/// 矩形の左上座標の取得、設定
		/// </summary>
		public Point Location
		{
			get { return new Point(_x1, _y1); }
			set
			{
				_x1 = value.X;
				_y1 = value.Y;
			}
		}

		/// <summary>
		/// Rectangle への暗黙的キャスト
		/// </summary>
		/// <param name="rc"></param>
		/// <returns></returns>
		public static implicit operator Rectangle(RectangleV rc)
		{
			return new Rectangle(rc.Location, rc.Size);
		}

		/// <summary>
		/// RectangleF への暗黙的キャスト
		/// </summary>
		/// <param name="rc"></param>
		/// <returns></returns>
		public static implicit operator RectangleF(RectangleV rc)
		{
			return new RectangleF(rc.Location, rc.Size);
		}

		/// <summary>
		/// 四角形の四隅の座標を左上から時計回りに 4 個の Point 配列で取得
		/// </summary>
		/// <param name="rect">四角形</param>
		/// <returns></returns>
		public Point[] Vertexes
		{
			get
			{
				// 四隅の座標
				Point[] points =
				{
					new Point(X1, Y1),
					new Point(X2, Y1),
					new Point(X2, Y2),
					new Point(X1, Y2)
				};

				return points;
			}
		}

		/// <summary>
		/// 頂点座標を指定するコンストラクタ
		/// </summary>
		/// <param name="x1">左上 X 座標</param>
		/// <param name="y1">左上 Y 座標</param>
		/// <param name="x2">右下 X 座標</param>
		/// <param name="y2">右下 Y 座標</param>
		public RectangleV(int x1, int y1, int x2, int y2)
		{
			_x1 = x1;
			_y1 = y1;
			_x2 = x2;
			_y2 = y2;
		}

		/// <summary>
		/// 左上位置とサイズを指定するコンストラクタ
		/// </summary>
		/// <param name="location">矩形の左上座標</param>
		/// <param name="size">矩形サイズ</param>
		public RectangleV(Point location, Size size)
		{
			_x1 = location.X;
			_y1 = location.Y;
			_x2 = _x1 + size.Width - 1;
			_y2 = _y1 + size.Height - 1;
		}

		/// <summary>
		/// Rectangle で初期化するコンストラクタ
		/// </summary>
		/// <param name="rect">Rectangle型の四角形</param>
		public RectangleV(Rectangle rect)
		{
			_x1 = rect.Left;
			_y1 = rect.Top;
			_x2 = _x1 + rect.Width - 1;
			_y2 = _y1 + rect.Height - 1;
		}

		/// <summary>
		/// RectangleF で初期化するコンストラクタ
		/// </summary>
		/// <param name="rect">RectangleF型の四角形</param>
		public RectangleV(RectangleF rect)
		{
			_x1 = (int)rect.Left;
			_y1 = (int)rect.Top;
			_x2 = (int)_x1 + (int)(rect.Width - 1);
			_y2 = (int)_y1 + (int)(rect.Height - 1);
		}

		/// <summary>
		/// 現在の四角形を dx、dyで移動した四角形を返す。
		/// </summary>
		/// <param name="dx">X 方向の移動量</param>
		/// <param name="dy">Y 方向の移動量</param>
		public RectangleV Move(int dx, int dy)
		{
			RectangleV tmp = this;

			tmp._x1 += dx;
			tmp._y1 += dy;
			tmp._x2 += dx;
			tmp._y2 += dy;

			return tmp;
		}

		/// <summary>
		/// 現在の四角形を dx、dyで移動した四角形を返す。
		/// その際に size で移動範囲を制限
		/// </summary>
		/// <param name="dx">X 方向移動量</param>
		/// <param name="dy">Y 方向移動量</param>
		/// <param name="size">制限範囲</param>
		/// <returns></returns>
		public RectangleV Move(int dx, int dy, Size size)
		{
			RectangleV tmp = Move(dx, dy);

			if(tmp.X1 < 0)
			{
				tmp.X2 = Width - 1;
				tmp.X1 = 0;
			}
			if(tmp.Y1 < 0)
			{
				tmp.Y2 = Height - 1;
				tmp.Y1 = 0;
			}
			if(tmp.X2 >= size.Width)
			{
				tmp.X1 = size.Width - Width;
				tmp.X2 = size.Width - 1;
			}
			if(tmp.Y2 >= size.Height)
			{
				tmp.Y1 = size.Height - Height;
				tmp.Y2 = size.Height - 1;
			}

			return tmp;
		}

		/// <summary>
		/// ポイント pt が、四角形内であるかを調べます
		/// </summary>
		/// <param name="pt">調べたいポイント</param>
		/// <returns></returns>
		public bool PtInRect(Point pt)
		{
			if(pt.X < X1)
				return false;

			if(pt.X >= X1 + Width)
				return false;

			if(pt.Y < Y1)
				return false;

			if(pt.Y >= Y1 + Height)
				return false;

			return true;
		}

		/// <summary>
		/// ポイント pt が、四角形内であるかを調べます（マージン付き）
		/// </summary>
		/// <param name="pt">調べたいポイント</param>
		/// <param name="margin">マージン</param>
		/// <returns></returns>
		public bool PtInRect(Point pt, int margin)
		{
			RectangleV testRect = this;

			testRect.X1 += margin;
			testRect.Y1 += margin;
			testRect.X2 -= margin;
			testRect.Y2 -= margin;

			return testRect.PtInRect(pt);
		}
	}
}
