﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Drawing;
using System.Windows.Forms;

namespace NauPlaying
{
	/// <summary>
	/// 領域部位識別列挙子
	/// </summary>
	public enum RegionalType
	{
		/// <summary>
		/// なし
		/// </summary>
		None,

		/// <summary>
		/// 境域内
		/// </summary>
		Inner,

		/// <summary>
		/// 左上
		/// </summary>
		TopLeft,
		/// <summary>
		/// 右上
		/// </summary>
		TopRight,
		/// <summary>
		/// 右下
		/// </summary>
		BottomRight,
		/// <summary>
		/// 左下
		/// </summary>
		BottomLeft,

		/// <summary>
		/// 上辺
		/// </summary>
		TopEdge,
		/// <summary>
		/// 右辺
		/// </summary>
		RightEdge,
		/// <summary>
		/// 下辺
		/// </summary>
		BottomEdge,
		/// <summary>
		/// 左辺
		/// </summary>
		LeftEdge,
	}

	/// <summary>
	/// 四角形領域操作クラス
	/// </summary>
	public class RegionRectEditor : IDisposable
	{
		/// <summary>
		/// 描画用ビットマップ
		/// </summary>
		private Bitmap _drawingBmp;
		/// <summary>
		/// 四角形領域が描画されたビットマップを取得
		/// </summary>
		public Bitmap RegionRectBtmap
		{
			get { return _drawingBmp; }
		}

		/// <summary>
		/// 移動中であるかのフラグ
		/// </summary>
		private bool _isMoving;

		/// <summary>
		/// 移動中であるかを調べる。または設定する。
		/// </summary>
		public bool IsMoving
		{
			set
			{
				if(value)
				{
					_movingRegionRect = _regionRect;
				}
				else
				{
					_regionRect = _movingRegionRect;
					_movingRegionRect = new RectangleV();
				}

				_isMoving = value;
			}
			get
			{
				return _isMoving;
			}
		}

		/// <summary>
		/// 四角形領域
		/// </summary>
		private RectangleV _regionRect;
		/// <summary>
		/// 四角形領域の設定、取得
		/// </summary>
		public Rectangle RegionRect
		{
			set { _regionRect = new RectangleV(value); }
			get { return _regionRect; }
		}

		/// <summary>
		/// 移動中の四角形領域
		/// </summary>
		private RectangleV _movingRegionRect;
		/// <summary>
		/// 移動中の四角形領域の取得。
		/// 移動中でない場合は不定。
		/// </summary>
		public RectangleV MovingRegionRect
		{
			get { return _movingRegionRect; }
		}

		/// <summary>
		/// サイズの取得
		/// </summary>
		public Size Size { private set; get; }

		/// <summary>
		/// 領域の最小サイズの設定、取得
		/// </summary>
		public Size MinimumRegionSize { set; get; }

		/// <summary>
		/// 頂点指示円の半径の設定、取得
		/// </summary>
		public int VertexCircleR { set; get; }
		/// <summary>
		/// 頂点や線がヒットしたと判定するマージンの設定、取得
		/// </summary>
		public int HitMargin { set; get; }

		/// <summary>
		/// 頂点指示円の描画色の設定、取得
		/// </summary>
		public Color VertexCircleFillColor { set; get; }
		/// <summary>
		/// 領域内の塗りつぶし色の設定、取得
		/// </summary>
		public Color RgnFillColor { set; get; }
		/// <summary>
		/// ライン描画の際のライン色の設定、取得
		/// </summary>
		public Color LinePenColor { set; get; }

		/// <summary>
		/// 正方形モードであるかの設定、取得
		/// </summary>
		public bool IsSquareMode { set; get; }

		/// <summary>
		/// 表示テキストの取得、設定
		/// </summary>
		public string Text { set; get; }
		/// <summary>
		/// 表示テキストフォントの取得、設定
		/// </summary>
		public Font Font { set; get; }
		/// <summary>
		/// テキスト色の取得、設定
		/// </summary>
		public Color ForeColor { set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="size"></param>
		public RegionRectEditor(Size size)
		{
			Size = size;
			MinimumRegionSize = new System.Drawing.Size(10, 10);

			_regionRect = new RectangleV(new Point(10, 10), new System.Drawing.Size(20, 20));

			VertexCircleR = 3;
			HitMargin = 3;

			VertexCircleFillColor = Color.FromArgb(255, 0, 0);
			RgnFillColor = Color.FromArgb(48, 0, 0, 255);
			LinePenColor = Color.FromArgb(0, 0, 0);

			IsSquareMode = false;

			Text = "";
			Font = new Font("MS UI Gothic", 12, FontStyle.Bold);
			ForeColor = Color.Black;

			Clear();
			UpdateRegionRect(_regionRect);
		}

		~RegionRectEditor()
		{
			Dispose(false);
		}

		/// <summary>
		/// IDisposable リソースの開放処理
		/// </summary>
		public void Dispose()
		{
			//マネージリソースおよびアンマネージリソースの解放
			this.Dispose(true);
			//ガベコレから、このオブジェクトのデストラクタを対象外とする
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if(_drawingBmp != null)
				_drawingBmp.Dispose();

			_drawingBmp = null;
		}

		/// <summary>
		/// 描画ビットマップクリア
		/// </summary>
		public void Clear()
		{
			if(_drawingBmp != null)
				_drawingBmp.Dispose();

			_drawingBmp = new Bitmap(Size.Width, Size.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
		}

		/// <summary>
		/// 四角形領域の変形
		/// </summary>
		/// <param name="type">操作する部位</param>
		/// <param name="dx">X 方向の変化量</param>
		/// <param name="dy">Y 方向の変化量</param>
		public void StretchRegionRect(RegionalType type, int dx, int dy)
		{
			_movingRegionRect = _regionRect;

			// X1 座標検査するデリゲート
			Action CheackX1 = () =>
			{
				if(_movingRegionRect.X1 < 0)
					_movingRegionRect.X1 = 0;
				if(_movingRegionRect.X1 > _movingRegionRect.X2 - MinimumRegionSize.Width)
					_movingRegionRect.X1 = _movingRegionRect.X2 - MinimumRegionSize.Width;
			};

			// Y1 座標検査するデリゲート
			Action CheackY1 = () =>
			{
				if(_movingRegionRect.Y1 < 0)
					_movingRegionRect.Y1 = 0;
				if(_movingRegionRect.Y1 > _movingRegionRect.Y2 - MinimumRegionSize.Height)
					_movingRegionRect.Y1 = _movingRegionRect.Y2 - MinimumRegionSize.Height;
			};

			// X2 座標検査するデリゲート
			Action CheackX2 = () =>
			{
				if(_movingRegionRect.X2 < _movingRegionRect.X1 + MinimumRegionSize.Width)
					_movingRegionRect.X2 = _movingRegionRect.X1 + MinimumRegionSize.Width;
				if(_movingRegionRect.X2 >= Size.Width)
					_movingRegionRect.X2 = Size.Width - 1;
			};

			// Y2 座標検査するデリゲート
			Action CheackY2 = () =>
			{
				if(_movingRegionRect.Y2 < _movingRegionRect.Y1 + MinimumRegionSize.Height)
					_movingRegionRect.Y2 = _movingRegionRect.Y1 + MinimumRegionSize.Height;
				if(_movingRegionRect.Y2 >= Size.Height)
					_movingRegionRect.Y2 = Size.Height - 1;
			};

			switch(type)
			{
			case RegionalType.Inner:
				_movingRegionRect = _movingRegionRect.Move(dx, dy, Size);
				break;

			case RegionalType.TopLeft:
				_movingRegionRect.X1 += dx;
				_movingRegionRect.Y1 += dy;

				CheackX1();
				CheackY1();

				if(IsSquareMode && _movingRegionRect.Width != _movingRegionRect.Height)
				{
					if(_movingRegionRect.Width < _movingRegionRect.Height)
						_movingRegionRect.Y1 = _movingRegionRect.Y2 - (_movingRegionRect.Width - 1);
					else
						_movingRegionRect.X1 = _movingRegionRect.X2 - (_movingRegionRect.Height - 1);
				}
				break;

			case RegionalType.TopRight:
				_movingRegionRect.X2 += dx;
				_movingRegionRect.Y1 += dy;

				CheackX2();
				CheackY1();

				if(IsSquareMode && _movingRegionRect.Width != _movingRegionRect.Height)
				{
					if(_movingRegionRect.Width < _movingRegionRect.Height)
						_movingRegionRect.Y1 = _movingRegionRect.Y2 - (_movingRegionRect.Width - 1);
					else
						_movingRegionRect.X2 = _movingRegionRect.X1 + (_movingRegionRect.Height - 1);
				}
				break;

			case RegionalType.BottomRight:
				_movingRegionRect.X2 += dx;
				_movingRegionRect.Y2 += dy;

				CheackX2();
				CheackY2();

				if(IsSquareMode && _movingRegionRect.Width != _movingRegionRect.Height)
				{
					if(_movingRegionRect.Width < _movingRegionRect.Height)
						_movingRegionRect.Y2 = _movingRegionRect.Y1 + (_movingRegionRect.Width - 1);
					else
						_movingRegionRect.X2 = _movingRegionRect.X1 + (_movingRegionRect.Height - 1);
				}
				break;

			case RegionalType.BottomLeft:
				_movingRegionRect.X1 += dx;
				_movingRegionRect.Y2 += dy;

				CheackX1();
				CheackY2();

				if(IsSquareMode && _movingRegionRect.Width != _movingRegionRect.Height)
				{
					if(_movingRegionRect.Width < _movingRegionRect.Height)
						_movingRegionRect.Y2 = _movingRegionRect.Y1 + (_movingRegionRect.Width - 1);
					else
						_movingRegionRect.X1 = _movingRegionRect.X2 - (_movingRegionRect.Height - 1);
				}
				break;

			case RegionalType.TopEdge:
				if(!IsSquareMode)
				{
					_movingRegionRect.Y1 += dy;
					CheackY1();
				}
				break;

			case RegionalType.RightEdge:
				if(!IsSquareMode)
				{
					_movingRegionRect.X2 += dx;
					CheackX2();
				}
				break;

			case RegionalType.BottomEdge:
				if(!IsSquareMode)
				{
					_movingRegionRect.Y2 += dy;
					CheackY2();
				}
				break;

			case RegionalType.LeftEdge:
				if(!IsSquareMode)
				{
					_movingRegionRect.X1 += dx;
					CheackX1();
				}
				break;
			}
		}

		/// <summary>
		/// 四角形領域描画更新。
		/// 移動中である場合（IsMoving == true）移動中の領域で描画更新されます。
		/// </summary>
		public void UpdateRegionRect()
		{
			if(IsMoving)
				UpdateRegionRect(MovingRegionRect);
			else
				UpdateRegionRect(_regionRect);
		}

		/// <summary>
		/// 四角形領域描画更新
		/// </summary>
		/// <param name="rect">四角形領域</param>
		private void UpdateRegionRect(RectangleV rect)
		{
			Clear();

			using(Graphics g = Graphics.FromImage(_drawingBmp))
			using(Brush pointBrush = new SolidBrush(VertexCircleFillColor))
			using(Brush rgnBrush = new SolidBrush(RgnFillColor))
			using(Pen pen = new Pen(LinePenColor))
			{
				// 点を描画するデリゲート
				Action<Point> DrawCirecle = (center) =>
				{
					Rectangle rc = new Rectangle(center.X - VertexCircleR, center.Y - VertexCircleR, VertexCircleR * 2 + 1, VertexCircleR * 2 + 1);

					g.FillEllipse(pointBrush, rc);
					g.DrawEllipse(pen, rc);
				};

				// 頂点座標の取得
				Point[] vertexes = rect.Vertexes;

				// 領域の描画
				g.FillRectangle(rgnBrush, rect);

				// 辺を描画
				pen.DashStyle = System.Drawing.Drawing2D.DashStyle.Dash;

				g.DrawLine(pen, vertexes[0], vertexes[1]);
				g.DrawLine(pen, vertexes[1], vertexes[2]);
				g.DrawLine(pen, vertexes[2], vertexes[3]);
				g.DrawLine(pen, vertexes[3], vertexes[0]);

				// 頂点を描画
				pen.DashStyle = System.Drawing.Drawing2D.DashStyle.Solid;

				foreach(Point pt in vertexes)
					DrawCirecle(pt);

				// 辺の中心に点を描画
				if(!IsSquareMode)
				{
					Point[] points = new Point[4];
					points[0] = new Point((rect.Right - rect.Left) / 2 + rect.Left, rect.Top);
					points[1] = new Point(rect.Right, (rect.Bottom - rect.Top) / 2 + rect.Top);
					points[2] = new Point((rect.Right - rect.Left) / 2 + rect.Left, rect.Bottom);
					points[3] = new Point(rect.Left, (rect.Bottom - rect.Top) / 2 + rect.Top);

					foreach(Point pt in points)
						DrawCirecle(pt);
				}

				// テキストの描画
				if(Text.Length > 0)
				{
					SizeF textSize = g.MeasureString(Text, Font);
					
					int x = (rect.Width / 2) - (int)(textSize.Width / 2) + rect.X1;
					int y = (rect.Height / 2) - (int)(textSize.Height / 2) + rect.Y1;

					using(Brush brush = new SolidBrush(ForeColor))
					{
						g.DrawString(Text, Font, brush, new PointF(x, y));
					}
				}
			}
		}

		/// <summary>
		/// ポイント pt で、領域のヒットテストを行う
		/// </summary>
		/// <param name="pt">調べたいポイント</param>
		/// <returns>ヒットテスト結果</returns>
		public RegionalType HitTest(Point pt)
		{
			return HitTest(pt, _regionRect);
		}

		/// <summary>
		/// ポイント pt で、領域 rect のヒットテストを行う
		/// </summary>
		/// <param name="pt">調べたいポイント</param>
		/// <param name="rect">調べる対象の領域</param>
		/// <returns>ヒットテスト結果</returns>
		private RegionalType HitTest(Point pt, RectangleV rect)
		{
			RegionalType hit;

			// 四隅?
			hit = HitTestVertexes(rect.Vertexes, VertexCircleR, pt, HitMargin);
			if(hit != RegionalType.None)
				return hit;

			// 辺？
			if(!IsSquareMode)
			{
				hit = HitTestEdges(rect, pt, HitMargin);
				if(hit != RegionalType.None)
					return hit;
			}

			// 領域内？
			if(rect.PtInRect(pt, HitMargin))
				return RegionalType.Inner;

			return RegionalType.None;
		}

		/// <summary>
		/// ポイント pt が、四角形領域のいずれかの頂点を指しているかを調べます
		/// </summary>
		/// <param name="points">四角形領域の四隅を格納した配列</param>
		/// <param name="r">判定半径</param>
		/// <param name="pt">調べたいポイント</param>
		/// <param name="margin">マージン</param>
		/// <returns>ヒットテスト結果</returns>
		private static RegionalType HitTestVertexes(Point[] points, int r, Point pt, int margin)
		{
			for(int i = 0 ; i < points.Length ; i++)
			{
				RectangleV testRect = new RectangleV(points[i].X - (r + margin), points[i].Y - (r + margin), points[i].X + (r + margin), points[i].Y + (r + margin));

				if(testRect.PtInRect(pt))
				{
					switch(i)
					{
					case 0:
						return RegionalType.TopLeft;

					case 1:
						return RegionalType.TopRight;

					case 2:
						return RegionalType.BottomRight;

					case 3:
						return RegionalType.BottomLeft;
					
					default:
						break;
					}
				}
			}

			return RegionalType.None;
		}

		/// <summary>
		/// ポイント pt が、四角形のいずれかの辺を指しているかを調べます
		/// </summary>
		/// <param name="rect">対象の四角形</param>
		/// <param name="pt">調べたいポイント</param>
		/// <param name="margin">マージン</param>
		/// <returns>ヒットテスト結果</returns>
		private static RegionalType HitTestEdges(RectangleV rect, Point pt, int margin)
		{
			// 上辺?
			if(pt.X >= rect.X1 && pt.X <= rect.X2 && pt.Y >= rect.Y1 - margin && pt.Y <= rect.Y1 + margin)
				return RegionalType.TopEdge;
			// 右辺?
			if(pt.X >= rect.X2 - margin && pt.X <= rect.X2 + margin && pt.Y >= rect.Y1 && pt.Y <= rect.Y2)
				return RegionalType.RightEdge;
			// 下辺?
			if(pt.X >= rect.X1 && pt.X <= rect.X2 && pt.Y >= rect.Y2 - margin && pt.Y <= rect.Y2 + margin)
				return RegionalType.BottomEdge;
			// 左辺?
			if(pt.X >= rect.X1 - margin && pt.X <= rect.X1 + margin && pt.Y >= rect.Y1 && pt.Y <= rect.Y2)
				return RegionalType.LeftEdge;

			return RegionalType.None;
		}
	}
}
