﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Drawing;

namespace NauPlaying.Skin
{
	/// <summary>
	/// アニメーションデータクラス
	/// </summary>
	public class AnimationData
	{
		/// <summary>
		/// 表示更新情報クラス
		/// </summary>
		public DrawingInfo DrawingInfo { set; get; }
		/// <summary>
		/// アニメーションが完了した場合true
		/// </summary>
		public bool IsFinish { set; get; }
		/// <summary>
		/// 描画の更新の必要がある場合に true
		/// </summary>
		public bool IsRedraw { set; get; }

		/// <summary>
		/// 標準のコンストラクタ
		/// </summary>
		public AnimationData()
		{
			DrawingInfo = null;
			IsFinish = false;
			IsRedraw = true;
		}
	}

	/// <summary>
	/// アニメーションクラスの基底クラス
	/// </summary>
	public abstract class Animation
	{
		/// <summary>
		/// 現在のアニメーションが完了しているかを調べる
		/// </summary>
		public bool IsFinish { protected set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		public Animation()
		{
			IsFinish = false;
		}

		/// <summary>
		/// アニメーションの実行
		/// </summary>
		/// <returns></returns>
		public abstract void Execute(AnimationData e);
	}

	/// <summary>
	/// フェードアニメーションクラス
	/// </summary>
	public class FadeAnimation : Animation
	{
		/// <summary>
		/// true でフェードイン、flase でフェードアウト
		/// </summary>
		public bool IsFadein { private set; get; }

		/// <summary>
		/// 現在のアルファ
		/// </summary>
		float _currentAlpha;
		/// <summary>
		/// アルファの更新量
		/// </summary>
		float _deltaAlpha;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="total">フェード完了までの時間</param>
		/// <param name="step">タイマーのステップ時間</param>
		/// <param name="fadein">true でフェードイン、flase でフェードアウト</param>
		public FadeAnimation(int total, int step, bool fadein)
		{
			IsFadein = fadein;

			// 終了までのカウント数を計算
			int totalCount = total / step;

			if(totalCount > 0)
			{
				if(IsFadein)
				{
					_currentAlpha = 0.0f;
					_deltaAlpha = 1.0f / totalCount;
				}
				else
				{
					_currentAlpha = 1.0f;
					_deltaAlpha = - 1.0f / totalCount;
				}
				
			}
			else
			{
				_currentAlpha = 0.0f;
				_deltaAlpha = 0.0f;

				IsFinish = true;
			}

			
		}

		/// <summary>
		/// アニメーションの実行
		/// </summary>
		/// <returns></returns>
		public override void Execute(AnimationData e)
		{
			if(IsFinish)
			{
				e.DrawingInfo.Alpha = 0.0f;

				IsFinish = true;
				e.IsFinish = true;
				e.IsRedraw = false;

				return;
			}

			if(IsFadein)
			{
				if(_currentAlpha >= 1.0f)
				{
					IsFinish = true;
					e.IsFinish = true;

					e.DrawingInfo.Alpha = 1.0f;
				}
				else
					e.DrawingInfo.Alpha = _currentAlpha;
			}
			else
			{
				if(_currentAlpha <= 0.0f)
				{
					IsFinish = true;
					e.IsFinish = true;

					e.DrawingInfo.Alpha = 0.0f;
				}
				else
					e.DrawingInfo.Alpha = _currentAlpha;
			}

			_currentAlpha += _deltaAlpha;
		}
	}

	
	/// <summary>
	/// 停止アニメーションクラス
	/// </summary>
	public class StopAnimation : Animation
	{
		/// <summary>
		/// 停止時間
		/// </summary>
		public int StopTime { private set; get; }

		/// <summary>
		/// 停止総カウント
		/// </summary>
		int _totalCount;
		/// <summary>
		/// 現在のカウント
		/// </summary>
		int _currentCount;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="stop">停止時間[ms]</param>
		/// <param name="step">タイマーのステップ時間[ms]</param>
		public StopAnimation(int stop, int step)
		{
			StopTime = stop;
			_totalCount = stop / step;
			_currentCount = 0;

			if(_totalCount == 0)
				IsFinish = true;
		}

		/// <summary>
		/// アニメーションの実行
		/// </summary>
		/// <returns></returns>
		public override void Execute(AnimationData e)
		{
			if(IsFinish)
			{
				IsFinish = true;
				e.IsFinish = true;
				e.IsRedraw = false;

				return;
			}

			if(_currentCount >= _totalCount)
			{
				IsFinish = true;
				e.IsFinish = true;
				e.IsRedraw = false;
			}
			else
			{
				e.IsRedraw = false;
			}

			_currentCount++;
		}
	}

	/// <summary>
	/// 固定アルファアニメーションクラス
	/// </summary>
	public class ConstAlphaAnimation : Animation
	{
		/// <summary>
		/// 適応するアルファ
		/// </summary>
		public float Alpha { private set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="alpha">適応するアルファ</param>
		public ConstAlphaAnimation(float alpha)
		{
			Alpha = alpha;
		}

		/// <summary>
		/// アニメーションの実行
		/// </summary>
		/// <returns></returns>
		public override void Execute(AnimationData e)
		{
			if(IsFinish)
			{
				IsFinish = true;
				e.IsFinish = true;
				e.IsRedraw = false;

				return;
			}

			e.DrawingInfo.Alpha = Alpha;

			IsFinish = true;
			e.IsFinish = true;
		}
	}

	/// <summary>
	/// 位置変更アニメーションクラス
	/// （位置の変更のみを行う）
	/// </summary>
	public class ChangeLocationAnimation : Animation
	{
		/// <summary>
		/// 適応する位置
		/// </summary>
		public Point Location { private set; get; }

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="alpha">適応するアルファ</param>
		public ChangeLocationAnimation(Point location)
		{
			Location = location;
		}

		/// <summary>
		/// アニメーションの実行
		/// </summary>
		/// <returns></returns>
		public override void Execute(AnimationData e)
		{
			if(IsFinish)
			{
				IsFinish = true;
				e.IsFinish = true;
				e.IsRedraw = false;

				return;
			}

			e.DrawingInfo.Location = Location;

			IsFinish = true;
			e.IsFinish = true;
			e.IsRedraw = false;
		}
	}
	
	/// <summary>
	/// スクロールアニメーションクラス
	/// </summary>
	public class ScrollAnimation : Animation
	{
		/// <summary>
		/// 表示領域幅
		/// </summary>
		public int ViewWidth { private set; get; }
		/// <summary>
		/// テキスト全体の描画幅
		/// </summary>
		public int TextWidth { private set; get; }
		/// <summary>
		/// 1スクロールで変化する移動量
		/// </summary>
		public int PixelStep { private set; get; }
		/// <summary>
		/// true で前進、flase で後進スクロール
		/// </summary>
		public bool IsFowardScrool { private set; get; }
		/// <summary>
		/// スクロール加速度
		/// </summary>
		public float ScrollAcceleration { private set; get; }
		/// <summary>
		/// スクロールに要するカウント
		/// </summary>
		public int ScrollCount { private set; get; }

		/// <summary>
		/// 現在の表示位置
		/// </summary>
		int _currentPos;
		/// <summary>
		/// 開始表示位置
		/// </summary>
		int _beginPos;
		/// <summary>
		/// 終了表示位置
		/// </summary>
		int _endPos;
		/// <summary>
		/// 表示位置の更新量
		/// </summary>
		int _deltaPos;
		/// <summary>
		/// 現在のカウント
		/// </summary>
		int _count;
		/// <summary>
		/// <summary>
		/// 加速するかの設定
		/// </summary>
		bool _enableAcceleration;
		/// <summary>
		/// 最大化速度
		/// </summary>
		float _maxDeltaPos;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="viewWidth">表示領域幅</param>
		/// <param name="textWidth">テキスト全体の描画幅</param>
		/// <param name="pixelStep">1スクロールで変化する移動量</param>
		/// <param name="acc">スクロール加速度</param>
		/// <param name="foward">true で前進、flase で後進スクロール</param>
		public ScrollAnimation(int viewWidth, int textWidth, int pixelStep, float acc, bool foward)
		{
			ViewWidth = viewWidth;
			TextWidth = textWidth;
			PixelStep = pixelStep;
			IsFowardScrool = foward;
			ScrollAcceleration = acc;

			_count = 0;

			if(TextWidth > ViewWidth)
			{
				// スクロールに必要なカウント数の計算
				int scroll_dist = textWidth - viewWidth;
				if(scroll_dist <= 0)
					ScrollCount = 0;
				else
					ScrollCount = scroll_dist / pixelStep;

				if(IsFowardScrool)
				{
					_beginPos = 0;
					_endPos = TextWidth - ViewWidth;
					_deltaPos = PixelStep;
					_currentPos = _beginPos;
				}
				else
				{
					_beginPos = TextWidth - ViewWidth;
					_endPos = 0;
					_deltaPos = -PixelStep;
					_currentPos = _beginPos;
				}
			}
			else
			{
				_currentPos = 0;
				_beginPos = 0;
				_endPos = 0;
				_deltaPos = 0;

				IsFinish = true;
			}

			// 加速?
			if(ScrollAcceleration > 0)
			{
				_maxDeltaPos = _deltaPos * 10;
				_enableAcceleration = true;
			}
			else
			{
				_maxDeltaPos = _deltaPos;
				_enableAcceleration = false;
			}
		}

		/// <summary>
		/// アニメーションの実行
		/// </summary>
		/// <returns></returns>
		public override void Execute(AnimationData e)
		{
			if(IsFinish)
			{
				e.DrawingInfo.TextSrcPos = _endPos;

				IsFinish = true;
				e.IsFinish = true;
				e.IsRedraw = false;

				return;
			}

			if(IsFowardScrool)
			{
				if(_currentPos >= _endPos)
				{
					IsFinish = true;
					e.IsFinish = true;

					e.DrawingInfo.TextSrcPos = _endPos;
				}
				else
					e.DrawingInfo.TextSrcPos = _currentPos;
			}
			else
			{
				if(_currentPos <= _endPos)
				{
					IsFinish = true;
					e.IsFinish = true;

					e.DrawingInfo.TextSrcPos = _endPos;
				}
				else
					e.DrawingInfo.TextSrcPos = _currentPos;
			}

			// 加速描画での表示位置更新
			if(_enableAcceleration)
			{
				float newDelataPos = _count * ScrollAcceleration;

				if(IsFowardScrool)
				{
					if(newDelataPos > _maxDeltaPos)
						newDelataPos = _maxDeltaPos;
				}
				else
				{
					newDelataPos = -newDelataPos;

					if(newDelataPos < _maxDeltaPos)
						newDelataPos = _maxDeltaPos;
				}

				_currentPos += (int)newDelataPos;
			}
			// 等速描画での表示位置更新
			else
			{
				_currentPos += _deltaPos;
			}

			_count++;
		}
	}

}
