﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Drawing;

namespace NauPlaying.Skin
{
	/// <summary>
	/// スキン描画クラス
	/// </summary>
	public class SkinRender : IDisposable
	{
		/// <summary>
		/// スキン設定の取得
		/// （この設定は Initialise されたときにディープコピーされたもの）
		/// </summary>
		private SkinSetting _skinSetting;
		/// <summary>
		/// アプリケーション設定の取得
		/// （この設定は Initialise されたときにディープコピーされたもの）
		/// </summary>
		private AppSetting _appSetting;

		/// <summary>
		/// テキストの描画サイズ
		/// </summary>
		public Size TextDrawingSize
		{
			get { return _textBmp.Size; }
		}

		/// <summary>
		/// テキストのスクロールが必要な場合 true
		/// </summary>
		public bool IsScrollText
		{
			get { return _skinSetting.TextDispRect.Width < TextDrawingSize.Width; }
		}

		/// <summary>
		/// 描画の際に設定された描画設定
		/// (再描画の際に設定される描画情報)
		/// </summary>
		private DrawingInfo _drawInfo;

		/// <summary>
		/// レイヤー用ビットマップ
		/// </summary>
		private Bitmap _layerBmp;

		/// <summary>
		/// ベースビットマップ（土台となる）
		/// </summary>
		private Bitmap _baseBmp;
		/// <summary>
		/// バックグランドイメージ格納用ビットマップ
		/// </summary>
		private Bitmap _backgroundBmp;
		/// <summary>
		/// テキスト格納用ビットマップ
		/// </summary>
		private Bitmap _textBmp;
		/// <summary>
		/// アートワーク格納用ビットマップ
		/// </summary>
		private Bitmap _artworkBmp;

		/// <summary>
		/// 標準コンストラクタ
		/// </summary>
		public SkinRender()
		{
			_layerBmp = null;

			_baseBmp = null;
			_backgroundBmp = null;
			_textBmp = null;
			_artworkBmp = null;
		}

		~SkinRender()
		{
			Dispose(false);
		}

		/// <summary>
		/// IDisposable リソースの開放処理
		/// </summary>
		public void Dispose()
		{
			//マネージリソースおよびアンマネージリソースの解放
			this.Dispose(true);
			//ガベコレから、このオブジェクトのデストラクタを対象外とする
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if(_layerBmp != null)
				_layerBmp.Dispose();
			_layerBmp = null;

			if(_baseBmp != null)
				_baseBmp.Dispose();
			_baseBmp = null;

			if(_backgroundBmp != null)
				_backgroundBmp.Dispose();
			_backgroundBmp = null;

			if(_textBmp != null)
				_textBmp.Dispose();
			_textBmp = null;

			if(_artworkBmp != null)
				_artworkBmp.Dispose();
			_artworkBmp = null;
		}

		/// <summary>
		/// レイヤードウインドウに貼り付けるためのレイヤービットマップを取得
		/// </summary>
		public Bitmap LayerBitmap
		{
			get { return _layerBmp; }
		}

		/// <summary>
		/// アートワークが存在しない場合のイメージを取得
		/// </summary>
		public static Bitmap NoneArtworkImage
		{
			get
			{
				return CreateNoneArtworkImage(new Size(256, 256));
			}
		}

		/// <summary>
		/// アートワークが存在しない場合のイメージを生成
		/// </summary>
		/// <param name="size">サイズ</param>
		/// <returns></returns>
		public static Bitmap CreateNoneArtworkImage(Size size)
		{
			return CreateBitmap(size, 128, Color.FromArgb(0, 0, 96));
		}

		/// <summary>
		/// スキンの初期化処理
		/// </summary>
		/// <param name="appSetting">アプリケーション設定（ディープコピーされる）</param>
		/// <param name="skinSetting">スキン設定（ディープコピーされる）</param>
		public void InitilizeBitmaps(AppSetting appSetting, SkinSetting skinSetting)
		{
			_skinSetting = skinSetting.DeepCopy();
			_appSetting = appSetting.DeepCopy();

			_drawInfo = new DrawingInfo();
			
			// この時点で可能な初期化を行う
			SkinSetting set = _skinSetting;

			Size layerSize = set.BoardSize;
			if(_appSetting.ZoomMaginification != 1)
			{
				layerSize.Height = (int)((float)layerSize.Height * _appSetting.ZoomMaginification);
				layerSize.Width = (int)((float)layerSize.Width * _appSetting.ZoomMaginification);
			}
			if(_layerBmp != null)
				_layerBmp.Dispose();
			
			_layerBmp = CreateBitmap(layerSize, 0, Color.Red);

			CreateBackground();
			ClearBaseBitmap();
			
			// アートワーク、テキストビットマップを生成
			if(_textBmp != null)
				_textBmp.Dispose();

			_textBmp = CreateBitmap(set.TextDispRect.Size, 0, Color.Red);

			if(_artworkBmp != null)
				_artworkBmp.Dispose();

			if(!set.EnableArtwork)
				_artworkBmp = CreateBitmap(set.ArtworkDispRect.Size, 0, Color.FromArgb(0, 0, 0));
			else
				_artworkBmp = null;
		}

		/// <summary>
		/// テキストを設定します
		/// </summary>
		/// <param name="text">表示するテキスト</param>
		public void SetText(string text)
		{
			AppSetting aset = _appSetting;
			SkinSetting sset = _skinSetting;

			// フォントの選択
			Font font;

			if(aset.UseSkinTextFont)
				font = sset.TextFont;
			else
				font = aset.TextFont;

			// 描画設定の選択
			int OutlineWidth;
			TextDrawingMode TextDrawingMode;
			Color TextColor1;
			Color TextColor2;

			if(aset.UseSkinTextDrawingSetting)
			{
				OutlineWidth = sset.OutlineWidth;
				TextDrawingMode = sset.TextDrawingMode;
				TextColor1 = sset.TextColor1;
				TextColor2 = sset.TextColor2;
			}
			else
			{
				OutlineWidth = aset.OutlineWidth;
				TextDrawingMode = aset.TextDrawingMode;
				TextColor1 = aset.TextColor1;
				TextColor2 = aset.TextColor2;
			}

			// 描画後のサイズを取得して再度生成
			Size textSize = new Size();
			using(Graphics g = Graphics.FromImage(_textBmp))
			{
				SizeF size = g.MeasureString(text, font);

				textSize.Width = (int)size.Width + 1;
				textSize.Height = (int)size.Height + 1;
			}

			// ノーマル以外の場合はアウトライン幅を考慮
			if(TextDrawingMode != TextDrawingMode.Normal)
			{
				textSize.Width += OutlineWidth;
				textSize.Height += OutlineWidth;
			}

			if(_textBmp != null) _textBmp.Dispose();
			_textBmp = CreateBitmap(new Size(textSize.Width, textSize.Height), 0, Color.Red);

			// 描画
			using(Graphics g = Graphics.FromImage(_textBmp))
			{
				g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
				g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

				float sizeInPixels = font.SizeInPoints * g.DpiY / 72;

				switch(TextDrawingMode)
				{
				case TextDrawingMode.Normal:
					using(SolidBrush brush = new SolidBrush(TextColor1))
					{
						g.DrawString(text, font, brush, new PointF());
					}
					break;

				case TextDrawingMode.Outline:
					using(Pen pen = new Pen(TextColor2, OutlineWidth))
					{
						System.Drawing.Drawing2D.GraphicsPath path = new System.Drawing.Drawing2D.GraphicsPath();
						path.AddString(text, font.FontFamily, (int)font.Style, sizeInPixels, new PointF(), new StringFormat());
						g.DrawPath(pen, path);
					}
					break;

				case TextDrawingMode.Border:
					using(SolidBrush brush = new SolidBrush(TextColor1))
					using(Pen pen = new Pen(TextColor2, OutlineWidth))
					{
						// アウトライン
						System.Drawing.Drawing2D.GraphicsPath path = new System.Drawing.Drawing2D.GraphicsPath();
						path.AddString(text, font.FontFamily, (int)font.Style, sizeInPixels, new PointF(), new StringFormat());
						g.DrawPath(pen, path);
						// 文字
						g.DrawString(text, font, brush, new PointF());
					}
					break;

				case TextDrawingMode.Shadow:
					using(SolidBrush brush1 = new SolidBrush(TextColor1))
					using(SolidBrush brush2 = new SolidBrush(TextColor2))
					{
						// 影
						g.DrawString(text, font, brush2, new PointF(OutlineWidth, OutlineWidth));
						// 文字
						g.DrawString(text, font, brush1, new PointF());
					}
					break;

				case TextDrawingMode.BorderAndShadow:
					using(SolidBrush brush1 = new SolidBrush(TextColor1))
					using(SolidBrush brush2 = new SolidBrush(TextColor2))
					using(Pen pen = new Pen(TextColor2, OutlineWidth))
					{
						// 影
						g.DrawString(text, font, brush2, new PointF(OutlineWidth, OutlineWidth));
						// アウトライン
						System.Drawing.Drawing2D.GraphicsPath path = new System.Drawing.Drawing2D.GraphicsPath();
						path.AddString(text, font.FontFamily, (int)font.Style, sizeInPixels, new PointF(), new StringFormat());
						g.DrawPath(pen, path);
						// 文字
						g.DrawString(text, font, brush1, new PointF());
					}
					break;
				}
			}
		}

		/// <summary>
		/// アートワークを設定します
		/// </summary>
		/// <param name="artwork">設定するアートワーク。null で表示のクリアのみを行う</param>
		public void SetArtwork(Bitmap artwork)
		{
			SkinSetting set = _skinSetting;

			if(!set.EnableArtwork)
				return;

			if(_artworkBmp != null)
				_artworkBmp.Dispose();
			
			_artworkBmp = CreateBitmap(new Size(set.ArtworkDispRect.Width, set.ArtworkDispRect.Height), 0, Color.Red);

			if(artwork == null)
				return;

			using(Graphics g = Graphics.FromImage(_artworkBmp))
			{
				// 正方ではない場合アスペクトを維持する
				Rectangle dest;
				if(artwork.Width != artwork.Height)
				{
					if(artwork.Width > artwork.Height)
					{
						float ratio = (float)artwork.Height / (float)artwork.Width;
						int height = (int)((float)_artworkBmp.Size.Height * ratio);

						int offset = (_artworkBmp.Size.Width - height) / 2;

						dest = new Rectangle(0, offset, _artworkBmp.Size.Width, height);
					}
					else
					{
						float ratio = (float)artwork.Width / (float)artwork.Height;
						int width = (int)((float)_artworkBmp.Size.Width * ratio);

						int offset = (_artworkBmp.Size.Height - width) / 2;

						dest = new Rectangle(offset, 0, width, _artworkBmp.Size.Height);
					}
				}
				else
					dest = new Rectangle(0, 0, _artworkBmp.Size.Width, _artworkBmp.Size.Height);

				g.DrawImage(artwork, dest);
			}
		}

		/// <summary>
		/// 表示を新しい描画設定でアップデート
		/// </summary>
		/// <param name="setting">新しい描画設定</param>
		public void UpdateBoard(DrawingInfo info)
		{
			_drawInfo = info.DeepCopy();
			UpdateBoard();
		}

		/// <summary>
		/// 表示を現在の描画設定でアップデート
		/// </summary>
		public void UpdateBoard()
		{
			SkinSetting set = _skinSetting;

			ClearBaseBitmap();

			using(Graphics g = Graphics.FromImage(_baseBmp))
			{
				// アートワークのコピー
				if(_skinSetting.EnableArtwork)
				{
					g.DrawImage(_artworkBmp, set.ArtworkDispRect.X, set.ArtworkDispRect.Y);
				}

				// テキストのコピー
				int textWidth = _textBmp.Size.Width - _drawInfo.TextSrcPos;
				int textHeight = _textBmp.Size.Height;
				int destWidth = set.TextDispRect.Width;
				int destHeight = set.TextDispRect.Height;
				int destHestOffset = (destHeight - textHeight) / 2;
				if(destHestOffset < 0)
					destHestOffset = 0;

				// 転送先サイズが小さい場合
				if(destWidth < textWidth)
				{
					Rectangle src = new Rectangle(_drawInfo.TextSrcPos, 0, destWidth, textHeight);

					g.DrawImage(_textBmp, set.TextDispRect.X, set.TextDispRect.Y + destHestOffset, src, GraphicsUnit.Pixel);
				}
				// 転送先サイズのほうが大きいか、同じ場合
				else
				{
					Rectangle src = new Rectangle(_drawInfo.TextSrcPos, 0, textWidth, textHeight);

					g.DrawImage(_textBmp, set.TextDispRect.X, set.TextDispRect.Y + destHestOffset, src, GraphicsUnit.Pixel);
				}
			}

			// ズームの結果、ビットマップサイズが異なる場合はレイヤーの際生成
			Size layerSize = set.BoardSize;
			if(_appSetting.ZoomMaginification != 1)
			{
				layerSize.Height = (int)((float)layerSize.Height * _appSetting.ZoomMaginification);
				layerSize.Width = (int)((float)layerSize.Width * _appSetting.ZoomMaginification);
			}
			if(_layerBmp != null)
				_layerBmp.Dispose();

			_layerBmp = CreateBitmap(layerSize, 0, Color.Red);

			// レイヤーにコピー
			using(Graphics g = Graphics.FromImage(_layerBmp))
			{
				System.Drawing.Imaging.ColorMatrix cm = new System.Drawing.Imaging.ColorMatrix();				
				cm.Matrix00 = 1;
				cm.Matrix11 = 1;
				cm.Matrix22 = 1;
				cm.Matrix33 = _drawInfo.Alpha * _appSetting.BoardAlpah;
				cm.Matrix44 = 1;

				System.Drawing.Imaging.ImageAttributes ia = new System.Drawing.Imaging.ImageAttributes();
				ia.SetColorMatrix(cm);

				Rectangle dst = new Rectangle(0, 0, _layerBmp.Width, _layerBmp.Height);
				g.DrawImage(_baseBmp, dst, 0, 0, _baseBmp.Width, _baseBmp.Height, GraphicsUnit.Pixel, ia);
			}
		}

		/// <summary>
		/// ベースビットマップを初期化
		/// </summary>
		private void ClearBaseBitmap()
		{
			if(_baseBmp != null)
				_baseBmp.Dispose();

			_baseBmp = new Bitmap(_backgroundBmp);
		}

		/// <summary>
		/// バックグランドイメージの生成
		/// </summary>
		private void CreateBackground()
		{
			SkinSetting set = _skinSetting;

			// バックグランド描画
			if(_backgroundBmp != null)
				_backgroundBmp.Dispose();

			_backgroundBmp = CreateGraduationImage(set.BoardSize, set.BackColor1, set.BackColor2, set.BackImageMode, set.GraduationFactor);

			// バックグランドビットマップ描画処理
			if(set.EnableBgImage)
			{
				try
				{
					using(Graphics g = Graphics.FromImage(_backgroundBmp))
					using(Bitmap img = new Bitmap(Generic.BkImageFilepath))
					{
						switch(set.BgTransparentMode)
						{
						case BgTransparentMode.OriginColor:
							img.MakeTransparent(img.GetPixel(0, 0));
							break;

						case BgTransparentMode.ConstColor:
							img.MakeTransparent(set.BgTransparentColor);
							break;
						}

						g.DrawImage(img, set.BgImageDispRect);
					}
				}
				catch(Exception)
				{
					set.EnableBgImage = false;
				}
			}
		}

		/// <summary>
		/// 垂直方向グラデーション描画
		/// </summary>
		/// <param name="bmp">描画するビットマップ</param>
		/// <param name="getColor">lineに応じた描画色を返すメソッドへのデリゲート</param>
		private static void DrawVerticalGraduationBitmap(Bitmap bmp, Func<int, Color> getColor)
		{
			using(Graphics g = Graphics.FromImage(bmp))
			{
				for(int line = 0 ; line < bmp.Height ; line++)
				{
					Color col = getColor(line);

					using(Pen pen = new Pen(col, 1))
					{
						g.DrawLine(pen, 0, line, bmp.Width - 1, line);
					}
				}
			}
		}

		/// <summary>
		/// 水平方向グラデーション描画
		/// </summary>
		/// <param name="bmp">描画するビットマップ</param>
		/// <param name="getColor">lineに応じた描画色を返すメソッドへのデリゲート</param>
		private static void DrawHorizontalGraduationBitmap(Bitmap bmp, Func<int, Color> getColor)
		{
			using(Graphics g = Graphics.FromImage(bmp))
			{
				for(int line = 0 ; line < bmp.Width ; line++)
				{
					Color col = getColor(line);

					using(Pen pen = new Pen(col, 1))
					{
						g.DrawLine(pen, line, 0, line, bmp.Height - 1);
					}
				}
			}
		}

		/// <summary>
		/// 指定したアルファ、色でビットマップを作成する
		/// </summary>
		/// <param name="size">サイズ</param>
		/// <param name="alpha">不透明度</param>
		/// <param name="col">ウインドウ色</param>
		/// <returns></returns>
		private static Bitmap CreateBitmap(Size size, int alpha, Color col)
		{
			Bitmap bmp = new Bitmap(size.Width, size.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

			using(Graphics g = Graphics.FromImage(bmp))
			using(SolidBrush brush = new SolidBrush(Color.FromArgb(alpha, col)))
			{
				g.FillRectangle(brush, new Rectangle(0, 0, size.Width, size.Height));
			}

			return bmp;
		}

		/// <summary>
		/// 指定した色、アルファでビットマップをクリアする
		/// </summary>
		/// <param name="bmp">ビットマップ</param>
		/// <param name="alpha">不透明度</param>
		/// <param name="col">ウインドウ色</param>
		private static void ClearBitmap(Bitmap bmp, int alpha, Color col)
		{
			Size size = bmp.Size;

			using(Graphics g = Graphics.FromImage(bmp))
			using(SolidBrush brush = new SolidBrush(Color.FromArgb(alpha, col)))
			{
				g.FillRectangle(brush, new Rectangle(0, 0, size.Width, size.Height));
			}
		}

		/// <summary>
		/// BackImageMode でグラデーション処理を行ったビットマップを生成
		/// </summary>
		/// <param name="size">サイズ</param>
		/// <param name="col1">第1色</param>
		/// <param name="col2">第2色</param>
		/// <param name="mode">描画モード</param>
		/// <param name="factor">グラデーションファクター</param>
		/// <returns></returns>
		public static Bitmap CreateGraduationImage(Size size, Color col1, Color col2, BackImageMode mode, float factor)
		{
			// 色変化量の計算
			Func<int, ColorF> ColorDelta = (span) => new ColorF((float)(col2.R - col1.R) / span, (float)(col2.G - col1.G) / span, (float)(col2.B - col1.B) / span);

			// 塗りつぶし
			if(mode == BackImageMode.Solid)
				return CreateBitmap(size, 255, col1);
			// 透明
			if(mode == BackImageMode.Transparence)
				return CreateBitmap(size, 0, Color.Red);

			// 描画処理
			Bitmap bmp = CreateBitmap(size, 0, Color.Red);

			switch(mode)
			{
			case BackImageMode.VerticalGraduation1:
				{
					ColorF cd = ColorDelta(size.Height);
					DrawVerticalGraduationBitmap(bmp, (line) => ColorF.ConvColor(255f, col1.R + cd.R * line, col1.G + cd.G * line, col1.B + cd.B * line));
				}
				break;

			case BackImageMode.VerticalGraduation2:
				{
					ColorF cd = ColorDelta(size.Height);

					float ad = 255f / (size.Height / 2);

					int height = bmp.Height;
					int half_height = height / 2;

					DrawVerticalGraduationBitmap(bmp,
						(line) =>
						{
							float a = 0;

							if(line < half_height)
								a = 255f - ad * line * factor;
							else
								a = ad * (line - half_height) / factor;

							return ColorF.ConvColor(a, col1.R + cd.R * line, col1.G + cd.G * line, col1.B + cd.B * line);
						});
				}
				break;

			case BackImageMode.VerticalGraduation3:
				{
					ColorF cd = ColorDelta(size.Height);

					float ad = 255 / (size.Height / 2);

					int height = bmp.Height;
					int half_height = height / 2;

					DrawVerticalGraduationBitmap(bmp,
						(line) =>
						{
							float a = 0;

							if(line < half_height)
								a = ad * line * factor;
							else
								a = 255f - ad * (line - half_height) / factor;

							return ColorF.ConvColor(a, col1.R + cd.R * line, col1.G + cd.G * line, col1.B + cd.B * line);
						});
				}
				break;

			case BackImageMode.VerticalGraduation4:
				{
					float ad = 255f / size.Height;

					DrawVerticalGraduationBitmap(bmp, (line) => ColorF.ConvColor(255 - ad * line * factor, col1.R, col1.G, col1.B));
				}
				break;

			case BackImageMode.VerticalGraduation5:
				{
					float ad = 255f / size.Height;

					DrawVerticalGraduationBitmap(bmp, (line) => ColorF.ConvColor(ad * line * factor, col1.R, col1.G, col1.B));
				}
				break;

			case BackImageMode.HorizontalGraduation1:
				{
					ColorF cd = ColorDelta(size.Width);

					DrawHorizontalGraduationBitmap(bmp, (line) => ColorF.ConvColor(255f, col1.R + cd.R * line, col1.G + cd.G * line, col1.B + cd.B * line));
				}
				break;

			case BackImageMode.HorizontalGraduation2:
				{
					ColorF cd = ColorDelta(size.Width);

					float ad = 255f / (size.Width / 2);

					int width = bmp.Width;
					int half_width = width / 2;

					DrawHorizontalGraduationBitmap(bmp,
						(line) =>
						{
							float a = 0;

							if(line < half_width)
								a = ad * line * factor;
							else
								a = 255f - ad * (line - half_width) / factor;

							return ColorF.ConvColor(a, col1.R + cd.R * line, col1.G + cd.G * line, col1.B + cd.B * line);
						});
				}
				break;

			case BackImageMode.HorizontalGraduation3:
				{
					ColorF cd = ColorDelta(size.Width);

					float ad = 255f / (size.Width / 2);

					int width = bmp.Width;
					int half_width = width / 2;

					DrawHorizontalGraduationBitmap(bmp,
						(line) =>
						{
							float a = 0;

							if(line < half_width)
								a = 255f - ad * line * factor;
							else
								a = ad * (line - half_width) / factor;

							return ColorF.ConvColor(a, col1.R + cd.R * line, col1.G + cd.G * line, col1.B + cd.B * line);
						});
				}
				break;

			case BackImageMode.HorizontalGraduation4:
				{
					float ad = 255f / size.Width;

					DrawHorizontalGraduationBitmap(bmp, (line) => ColorF.ConvColor(255f - ad * line * factor, col1.R, col1.G, col1.B));
				}
				break;

			case BackImageMode.HorizontalGraduation5:
				{
					float ad = 255f / size.Width;

					DrawHorizontalGraduationBitmap(bmp, (line) => ColorF.ConvColor(ad * line * factor, col1.R, col1.G, col1.B));
				}
				break;
			}

			return bmp;
		}
	}
}
