﻿using System;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Win32SDK
{
	/// <summary>
	/// 低レベルキーボードフッククラス
	/// </summary>
	internal static class LLKeyboardHook
	{
		/// <summary>
		/// キーボードイベントパラメータクラス
		/// </summary>
		public class KeybordHookEventArgs : EventArgs
		{
			/// <summary>
			/// 仮想キーコード
			/// </summary>
			public UInt32 KeyCode { private set; get; }
			public UInt32 ScanCode { private set; get; }
			public UInt32 Flags { private set; get; }
			public UInt32 Time { private set; get; }
			/// <summary>
			/// イベントをキャンセルしたい場合は true
			/// </summary>
			public bool Cancel { set; get; }

			internal KeybordHookEventArgs(User32.KBDLLHOOKSTRUCT keyData)
			{
				KeyCode = keyData.vkCode;
				ScanCode = keyData.scanCode;
				Flags = keyData.flags;
				Time = keyData.time;
				Cancel = false;
			}
		}

		/// <summary>
		/// フックハンドル
		/// </summary>
		private static IntPtr _hookHandle = IntPtr.Zero;
		/// <summary>
		/// フックプロシージャ
		/// </summary>
		private static User32.LowLevelKeyboardProc _hookProc;

		/// <summary>
		/// キーボードフックが機能していれば true
		/// </summary>
		public static bool IsValidHook
		{
			get { return _hookHandle != IntPtr.Zero; }
		}

		/// <summary>
		/// キーダウン(WM_KEYDOWN)
		/// </summary>
		public static event EventHandler<KeybordHookEventArgs> OnKeyDown;
		/// <summary>
		/// キーアップイベント(WM_KEYUP)
		/// </summary>
		public static event EventHandler<KeybordHookEventArgs> OnKeyUp;
		/// <summary>
		/// システムキーダウン(WM_SYSKEYDOWN)
		/// </summary>
		public static event EventHandler<KeybordHookEventArgs> OnSysKeyDown;
		/// <summary>
		/// システムキーアップ(WM_SYSKEYUP)
		/// </summary>
		public static event EventHandler<KeybordHookEventArgs> OnSysKeyUp;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		static LLKeyboardHook()
		{
			_hookProc = new User32.LowLevelKeyboardProc(LowLevelKeyboardProc);

			System.Reflection.Module mod = typeof(LLKeyboardHook).Module;
			IntPtr handle = System.Runtime.InteropServices.Marshal.GetHINSTANCE(mod);

			_hookHandle = User32.SetWindowsHookEx(User32.WH_KEYBOARD_LL, _hookProc, handle, 0);

			if(_hookHandle != IntPtr.Zero)
				Debug.WriteLine("LLKeyboardHook : キーフック開始");
			else
				Debug.WriteLine("LLKeyboardHook : キーフック失敗 : " + Kernel32.FormatMessageFromLastError());

			AppDomain.CurrentDomain.DomainUnload += delegate
			{
				if(_hookHandle != IntPtr.Zero)
				{
					User32.UnhookWindowsHookEx(_hookHandle);
					Debug.WriteLine("LLKeyboardHook : キーフック停止");

					_hookHandle = IntPtr.Zero;
				}
				else
					Debug.WriteLine("LLKeyboardHook : キーフックは開始されていません。");
			};
		}

		/// <summary>
		/// フックプロシージャ
		/// </summary>
		/// <param name="nCode"></param>
		/// <param name="wParam"></param>
		/// <param name="lParam"></param>
		/// <returns></returns>
		static IntPtr LowLevelKeyboardProc(int nCode, IntPtr wParam, ref User32.KBDLLHOOKSTRUCT lParam)
		{
			bool cancel = false;

			if(nCode == User32.HC_ACTION)
			{
				KeybordHookEventArgs e = new KeybordHookEventArgs(lParam);

				switch(wParam.ToInt32())
				{
				case Win32.WM_KEYDOWN:
					if(OnKeyDown != null)
						OnKeyDown(null, e);
					break;

				case Win32.WM_KEYUP:
					if(OnKeyUp != null)
						OnKeyUp(null, e);
					break;

				case Win32.WM_SYSKEYDOWN:
					if(OnSysKeyDown != null)
						OnSysKeyDown(null, e);
					break;

				case Win32.WM_SYSKEYUP:
					if(OnSysKeyUp != null)
						OnSysKeyUp(null, e);
					break;
				}

				cancel = e.Cancel;
			}

			if(cancel)
				return (IntPtr)1;

			return User32.CallNextHookEx(_hookHandle, nCode, wParam, ref lParam);
		}
	}
}
