﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Collections.Specialized;

namespace PluginBase
{
	/// <summary>
	/// モニターイベント
	/// </summary>
	public enum AppMonitorEvent
	{
		/// <summary>
		/// なし
		/// </summary>
		None,
		/// <summary>
		/// 接続
		/// </summary>
		Connect,
		/// <summary>
		/// 切断
		/// </summary>
		Disconnect,
		/// <summary>
		/// 再生の停止
		/// </summary>
		StopPlayingTrack,
		/// <summary>
		/// 再生曲の変更
		/// </summary>
		ChangePlayingTrack,
		/// <summary>
		/// プラグインでエラーが発生した
		/// </summary>
		PluginErrorEvent,
	}

	/// <summary>
	/// プレーヤコントロールの種類
	/// </summary>
	public enum PlayerControl
	{
		/// <summary>
		/// 再生
		/// </summary>
		Play,
		/// <summary>
		/// 停止
		/// </summary>
		Stop,
		/// <summary>
		/// 次の曲
		/// </summary>
		Next,
		/// <summary>
		/// 前の曲
		/// </summary>
		Prev,
		/// <summary>
		/// 再生 <> 一時停止
		/// </summary>
		SwitchPlay,
	}

	/// <summary>
	/// アプリケーションモニター用プラグインインターフェース
	/// </summary>
	public interface IAppMonitorPlugin : IDisposable, IPlugin
	{
		#region /* プロパティ */

		/// <summary>
		/// サポートするプレーヤー名の取得
		/// </summary>
		string SupportPlayerName { get; }

		/// <summary>
		/// 現在再生中の曲情報（再生されていない、曲を取得できなかった場合は null）
		/// </summary>
		TrackInfo CurrentTrackInfo { get; }

		/// <summary>
		/// モニター中かを調べる
		/// </summary>
		bool IsBusy { get; }

		/// <summary>
		/// 現在再生中かを調べる
		/// </summary>
		bool IsPlaying { get; }

		/// <summary>
		/// 再生位置[s]の取得、設定
		/// </summary>
		int PlayerPosition { get; set; }

		/// <summary>
		/// 再生ボリュームを 0 ～ 100 [%] で取得、設定
		/// </summary>
		int SoundVolume { get; set; }

		/// <summary>
		/// リンクが確立されているかを調べる
		/// </summary>
		bool IsConnected { get; }

		/// <summary>
		/// 最後のエラーを返す（1回もエラーが発生していないときは null）
		/// </summary>
		Exception LastException { get; }

		#endregion

		#region /* 公開メソッド */

		/// <summary>
		/// モニターを開始します。正常にモニターを開始できた場合 true を返します。
		/// すでに開始されていた場合は false を返します。
		/// </summary>
		/// <param name="appMonitorInterval">監視タイマー間隔[ms](タイマーを使用しないのならば無視すればいい)</param>
		/// <param name="appStopWait">アプリケーションの終了を待機する時間[ms]（0で待機しない）</param>
		/// <returns>false 失敗、true 成功</returns>
		bool StartMonitor(int appMonitorInterval, int appStopWait);

		/// <summary>
		/// モニターを停止します。
		/// </summary>
		void StopMonitor();

		/// <summary>
		/// プレーヤを制御します
		/// </summary>
		/// <param name="type">制御タイプ</param>
		/// <returns>false 失敗、true 成功</returns>
		bool ControlPlayer(PlayerControl type);

		/// <summary>
		/// プラグインロード時に呼ばれる
		/// </summary>
		/// <param name="obj">未使用</param>
		void Init(object obj);
		/// <summary>
		/// プラグイン終了時に呼ばれる
		/// </summary>
		/// <param name="obj">未使用</param>
		void Exit(object obj);
		/// <summary>
		/// プラグインの設定を行う際に呼ばれる
		/// </summary>
		/// <param name="obj">未使用</param>
		void Config(object obj);

		/// <summary>
		/// モニターイベントをポップする
		/// </summary>
		/// <returns></returns>
		AppMonitorEvent PopAppMonitorEvent();

		#endregion
	}
}
