﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Threading;
using System.Runtime.InteropServices;
using System.IO;
using System.Drawing;
using System.Diagnostics;
using System.Collections.Specialized;
using System.Runtime.Remoting.Lifetime;

namespace PluginBase
{
	/// <summary>
	/// iTunes モニタープラグインクラス
	/// </summary>
	public class iTunesMonitorPlugin : MarshalByRefObject, IAppMonitorPlugin
	{
		/// <summary>
		/// リース期間の初期化
		/// </summary>
		/// <returns></returns>
		public override object InitializeLifetimeService()
		{
			return null;
		}

		/// <summary>
		/// itunes に接続済みだった場合は true(起動誤検出回避のヒステリシス動作用フラグ)
		/// </summary>
		private static bool _hysteresisFlag = false;
		/// <summary>
		/// iTunesApp COM
		/// </summary>
		private static iTunesLib.iTunesApp _app = null;
		/// <summary>
		/// iTunes 起動状況監視タイマー
		/// </summary>
		private static Timer _appMonitorTimer = null;
		/// <summary>
		/// iTunes起動状況監視タイマー間隔[ms]
		/// </summary>
		private int _appMonitorInterval;
		/// <summary>
		/// 多重実行防止フラグ
		/// </summary>
		private static bool _flag = false;

		/// <summary>
		/// 操作メソッドをスレッドセーフにするロック用オブジェクト
		/// </summary>
		private static object _operatorLockObj = new object();

		/// <summary>
		/// 添付ファイル操作ロック用オブジェクト
		/// （添付ファイルを使用する場合は、このオブジェクトでロックしてください）
		/// </summary>
		private static object TempFilepathLockObj = new object();
		/// <summary>
		/// 画像取得のための一時保管ファイルのファイルフルパスの取得
		/// </summary>
		private static readonly string TempFilepath = Path.GetTempFileName();

		/// <summary>
		/// イベントキュー
		/// </summary>
		private static EventQueue<AppMonitorEvent> _eventQueue = new EventQueue<AppMonitorEvent>();

		#region /* IAppMonitorPlugin 公開プロパティ */

		/// <summary>
		/// プラグインタイプの取得
		/// </summary>
		public PluginType MyPluginType { get { return PluginBase.PluginType.AppMonitor; } }
		/// <summary>
		/// プラグイン名の取得
		/// </summary>
		public string PluginName { get { return "iTunes App Monitor Plugin Ver1.00"; } }
		/// <summary>
		/// プラグインの著作権の取得
		/// </summary>
		public string Copyright { get { return "弱チルノ"; } }
		/// <summary>
		/// サポートするプレーヤー名の取得
		/// </summary>
		public string SupportPlayerName { get { return "iTunes"; } }

		/// <summary>
		/// 現在再生中の曲情報（再生されていない、曲を取得できなかった場合は null）
		/// </summary>
		public TrackInfo CurrentTrackInfo
		{
			get
			{
				if(!IsPlaying)
					return null;

				lock(_operatorLockObj)
				{
					TrackInfo track = null;
					try
					{
						track = CreateTrackInfo(_app.CurrentTrack);
					}
					catch(Exception)
					{
						track = null;
					}

					return track;
				}
			}
		}

		/// <summary>
		/// モニター中かを調べる
		/// </summary>
		public bool IsBusy
		{
			get
			{
				lock(_operatorLockObj)
				{
					return _appMonitorTimer != null;
				}
			}
		}

		/// <summary>
		/// 現在再生中かを調べる
		/// </summary>
		public bool IsPlaying
		{
			get
			{
				try
				{
					if(!IsConnected)
						return false;

					lock(_operatorLockObj)
					{
						return _app.PlayerState == iTunesLib.ITPlayerState.ITPlayerStatePlaying;
					}
				}
				catch(Exception)
				{
				}

				return false;
			}
		}

		/// <summary>
		/// 再生位置の取得、設定
		/// </summary>
		public int PlayerPosition
		{
			get
			{
				try
				{
					if(!IsConnected)
						return 0;

					lock(_operatorLockObj)
					{
						return _app.PlayerPosition;
					}
				}
				catch(Exception)
				{
				}

				return 0;
			}
			set
			{
				try
				{
					if(!IsConnected)
						return;

					lock(_operatorLockObj)
					{
						_app.PlayerPosition = value;
					}
				}
				catch(Exception)
				{
				}
			}
		}

		/// <summary>
		/// 再生ボリュームを 0 ～ 100 [%] で取得、設定
		/// </summary>
		public int SoundVolume
		{
			get
			{
				try
				{
					if(!IsConnected)
						return 0;

					lock(_operatorLockObj)
					{
						return _app.SoundVolume;
					}
				}
				catch(Exception)
				{
				}

				return 0;
			}
			set
			{
				try
				{
					if(!IsConnected)
						return;

					lock(_operatorLockObj)
					{
						_app.SoundVolume = value;
					}
				}
				catch(Exception)
				{
				}
			}
		}

		/// <summary>
		/// iTunes とのリンクが確立されているかを調べる
		/// </summary>
		public bool IsConnected
		{
			get
			{
				lock(_operatorLockObj)
				{
					return _app != null;
				}
			}
		}

		/// <summary>
		/// 最後のエラーを返す（1回もエラーが発生していないときは null）
		/// </summary>
		public Exception LastException { private set;  get; }

		#endregion

		#region /* IAppMonitorPlugin 公開メソッド */

		/// <summary>
		/// モニターを開始します。正常にモニターを開始できた場合 true を返します。
		/// すでに開始されていた場合は false を返します。
		/// </summary>
		/// <param name="appMonitorInterval">監視タイマー間隔[ms](タイマーを使用しないのならば無視すればいい)</param>
		/// <param name="appStopWait">アプリケーションの終了を待機する時間[ms]（0で待機しない）/param>
		/// <returns>false 失敗、true 成功</returns>
		public bool StartMonitor(int appMonitorInterval, int appStopWait)
		{
			if(IsBusy)
				return false;

			try
			{
				lock(_operatorLockObj)
				{
					_appMonitorInterval = appMonitorInterval;

					if(appStopWait > 0)
					{
						int count = 0;
						while(CheckRunningiTunesApp())
						{
							System.Threading.Thread.Sleep(500);
							count += 500;

							if(count >= appStopWait)
								break;
						}
					}

					_appMonitorTimer = new Timer(_appMonitorTimer_MonitorTask, null, 1000, _appMonitorInterval);
				}
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);

				return false;
			}

			return true;
		}

		/// <summary>
		/// モニターを停止します。
		/// </summary>
		public void StopMonitor()
		{
			try
			{
				if(!IsBusy)
					return;

				lock(_operatorLockObj)
				{
					DisposeTimer();
					DisconnectiTunes();
				}
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);
			}
			finally
			{
				_hysteresisFlag = false;
			}
		}

		/// <summary>
		/// iTunes を制御します
		/// </summary>
		/// <param name="type">制御タイプ</param>
		/// <returns></returns>
		public bool ControlPlayer(PlayerControl type)
		{
			try
			{
				if(!IsConnected)
					return false;

				lock(_operatorLockObj)
				{
					switch(type)
					{
					case PlayerControl.Play:
						_app.Play();
						break;

					case PlayerControl.Stop:
						_app.PlayPause();
						break;

					case PlayerControl.Next:
						_app.NextTrack();
						break;

					case PlayerControl.Prev:
						_app.PreviousTrack();
						break;

					case PlayerControl.SwitchPlay:
						if(_app.PlayerState == iTunesLib.ITPlayerState.ITPlayerStatePlaying)
							_app.PlayPause();
						else
							_app.Play();
						break;

					default:
						return false;
					}
				}
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);

				return false;
			}

			return true;
		}

		/// <summary>
		/// プラグインロード時に呼ばれる
		/// </summary>
		public void Init(object obj)
		{
		}

		/// <summary>
		/// プラグイン終了時に呼ばれる
		/// </summary>
		public void Exit(object obj)
		{
			lock(_operatorLockObj)
			{
				Dispose();
			}
		}

		/// <summary>
		/// プラグインの設定を行う際に呼ばれる
		/// </summary>
		public void Config(object obj)
		{
		}

		/// <summary>
		/// モニターイベントをポップする
		/// </summary>
		/// <returns></returns>
		public AppMonitorEvent PopAppMonitorEvent()
		{
			AppMonitorEvent appEvent;
			if(_eventQueue.Pop(out appEvent))
				return appEvent;

			return AppMonitorEvent.None;
		}

		#endregion

		/// <summary>
		/// コンストラクタ
		/// </summary>
		public iTunesMonitorPlugin()
		{
			_appMonitorInterval = 1000;
		}

		/// <summary>
		/// デストラクタ
		/// </summary>
		~iTunesMonitorPlugin()
		{
			Dispose(false);
		}

		/// <summary>
		/// IDisposable リソースの開放処理
		/// </summary>
		public void Dispose()
		{
			//マネージリソースおよびアンマネージリソースの解放
			this.Dispose(true);
			//ガベコレから、このオブジェクトのデストラクタを対象外とする
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			DisposeTimer();
			DisconnectiTunes();

			// テンプファイルの削除
			RemoveTempFile();
		}

		/// <summary>
		/// タイマーを破棄する（停止）
		/// </summary>
		private void DisposeTimer()
		{
			if(_appMonitorTimer != null)
			{
				_appMonitorTimer.Dispose();
				_appMonitorTimer = null;
			}
		}

		/// <summary>
		/// iTunes 起動状況監視タイマーイベント
		/// </summary>
		/// <param name="StateObj"></param>
		private void _appMonitorTimer_MonitorTask(object StateObj)
		{
			if(_flag)
				return;

			try
			{
				_flag = true;

				// すでに接続が確立されている場合は何もせずに終了
				if(IsConnected)
					return;

				// 起動されていた場合は iTunes の終了を確認するまで接続を再接続しない
				if(_hysteresisFlag)
				{
					if(!CheckRunningiTunesApp())
						_hysteresisFlag = false;
				}
				else
				{
					// iTunes が起動しているかを調べて起動していた場合はリンクを試みる
					if(CheckRunningiTunesApp())
					{
						ConnectiTunes();
						_hysteresisFlag = true;

						_eventQueue.Push(AppMonitorEvent.Connect);

						Debug.WriteLine("iTunes > Connected");
					}
				}
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);

				_hysteresisFlag = false;
			}
			finally
			{
				_flag = false;
			}
		}

		/// <summary>
		/// iTunesが起動されているかを調べる
		/// </summary>
		/// <returns></returns>
		private bool CheckRunningiTunesApp()
		{
			try
			{
				System.Diagnostics.Process[] ps = System.Diagnostics.Process.GetProcesses();

				foreach(System.Diagnostics.Process procs in ps)
				{
					if(procs.ProcessName == "iTunes")
						return true;
				}
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);
			}

			return false;
		}

		/// <summary>
		/// iTunes との接続を試みる
		/// </summary>
		/// <returns>成功した場合は true</returns>
		private void ConnectiTunes()
		{
			if(_app != null)
				return;

			try
			{
				_app = new iTunesLib.iTunesApp();

				_app.OnPlayerPlayEvent += _app_OnPlayerPlayEvent;
				_app.OnPlayerStopEvent += _app_OnPlayerStopEvent;
				_app.OnAboutToPromptUserToQuitEvent += _app_OnAboutToPromptUserToQuitEvent;
				_app.OnQuittingEvent += _app_OnQuittingEvent;
			}
			catch(Exception ex)
			{
				DisconnectiTunes();

				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);

				throw;
			}
		}

		/// <summary>
		/// iTunes との接続を切断
		/// </summary>
		private void DisconnectiTunes()
		{
			if(_app == null)
				return;

			try
			{
				int i = 0;
				do
				{
					i = Marshal.ReleaseComObject(_app);
				} while(i > 0);
			}
			finally
			{
				_app = null;
				Debug.WriteLine("iTunes > Disconnnected");
			}
		}

		/// <summary>
		/// iTunesが終了する前に呼ばれる
		/// </summary>
		private void _app_OnAboutToPromptUserToQuitEvent()
		{
			DisconnectiTunes();

			_eventQueue.Push(AppMonitorEvent.Disconnect);
		}

		/// <summary>
		/// iTunes が終了したときに呼ばれる
		/// </summary>
		private void _app_OnQuittingEvent()
		{
			DisconnectiTunes();
		}

		/// <summary>
		/// iTunesで再生が開始された時に呼ばれる
		/// </summary>
		/// <param name="_track"></param>
		private void _app_OnPlayerPlayEvent(object _track)
		{
			try
			{
				iTunesLib.IITTrack itrack = (iTunesLib.IITTrack)_track;

				if(itrack != null)
				{
					_eventQueue.Push(AppMonitorEvent.ChangePlayingTrack);
				}
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);
			}
		}

		/// <summary>
		/// iTunesで再生が停止された時に呼ばれる
		/// </summary>
		/// <param name="_track"></param>
		private void _app_OnPlayerStopEvent(object _track)
		{
			try
			{
				_eventQueue.Push(AppMonitorEvent.StopPlayingTrack);
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);
			}
		}

		/// <summary>
		/// 一時ファイルを削除します
		/// </summary>
		private static void RemoveTempFile()
		{
			if(!System.Threading.Monitor.TryEnter(TempFilepathLockObj, 5000))
			{
				System.Windows.Forms.MessageBox.Show("一時ファイルのロックに失敗しました。\n＊次回正常終了時に削除されます。", "iTunes Monitor Plugin", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Exclamation);
				return;
			}

			try
			{
				if(File.Exists(TempFilepath))
					File.Delete(TempFilepath);
			}
			catch(Exception)
			{
				System.Windows.Forms.MessageBox.Show("一時ファイルの削除に失敗しました。\n＊次回正常終了時に削除されます。", "iTunes Monitor Plugin", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Exclamation);
			}
			finally
			{
				System.Threading.Monitor.Exit(TempFilepathLockObj);
			}
		}

		/// <summary>
		/// iTunesLib.IITTrack で TrackInfo を生成
		/// </summary>
		/// <param name="itrack"></param>
		private TrackInfo CreateTrackInfo(iTunesLib.IITTrack itrack)
		{
			try
			{
				TrackInfo info = new TrackInfo();
				
				info.Name = itrack.Name;
				info.Album = itrack.Album;
				info.Artist = itrack.Artist;
				info.TrackNo = itrack.TrackNumber.ToString();
				info.ArtworkFilepath = "";
				info.Duration = itrack.Duration;

				iTunesLib.IITArtworkCollection artworks = itrack.Artwork;
				if(artworks.Count > 0)
				{
					iTunesLib.IITArtwork artwork = artworks[1];

					iTunesLib.ITArtworkFormat format = artwork.Format;
					string ArtworkExtension = "";

					switch(format)
					{
					case iTunesLib.ITArtworkFormat.ITArtworkFormatJPEG:
						ArtworkExtension = ".jpg";
						break;
					case iTunesLib.ITArtworkFormat.ITArtworkFormatPNG:
						ArtworkExtension = ".png";
						break;
					case iTunesLib.ITArtworkFormat.ITArtworkFormatBMP:
						ArtworkExtension = ".bmp";
						break;
					}

					if(ArtworkExtension.Length > 0)
					{
						try
						{
							// iTunes のアートワーク書き込み中に、他のスレッドからさらに書き込まれるのを防ぐロック
							lock(TempFilepathLockObj)
							{
								artwork.SaveArtworkToFile(TempFilepath);
							}

							info.ArtworkFilepath = TempFilepath;
						}
						catch(Exception e)
						{
							System.Diagnostics.Debug.WriteLine("TrackInfo(iTunesLib.IITTrack itrack) Exception! > " + e.Message);

							info.ArtworkFilepath = "";
						}
					}
				}

				return info;
			}
			catch(Exception ex)
			{
				LastException = ex;
				_eventQueue.Push(AppMonitorEvent.PluginErrorEvent);
			}

			return null;
		}
	}
}
